# -*- coding: utf-8 -*-
from neumann.numint import GaussPoints as Gauss

from polymesh.cells import H8 as HexaHedron

from ..model.solid3d import Solid3d

from .elem import FiniteElement
from .meta import ABCFiniteElement as ABC


class H8(ABC, Solid3d, HexaHedron, FiniteElement):

    qrule = 'selective'
    quadrature = {
        'full': Gauss(2, 2, 2),
        'selective': {
            (0, 1, 2): 'full',
            (3, 4, 5): 'reduced'
        },
        'reduced': Gauss(1, 1, 1)
    }



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange
__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def approx_element_solution_bulk(dofsol: ndarray, A: ndarray):
    """
    dofsol (nE, nRHS, nDOF * nNE)
    A (nE, nP, nX, nDOF * nNE)
    ---
    (nE, nRHS, nP, nX)
    """
    nE, nP, nX = A.shape[:3]
    nRHS = dofsol.shape[1]
    res = np.zeros((nE, nRHS, nP, nX), dtype=dofsol.dtype)
    for i in prange(nE):
        for j in prange(nP):
            for k in prange(nRHS):
                res[i, k, j, :] = A[i, j] @ dofsol[i, k]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def calculate_internal_forces_bulk(strains: ndarray, D: ndarray):
    """
    strain (nE, nRHS, nP, nX)
    D (nE, nX, nX)
    ---
    (nE, nRHS, nP, nX)
    """
    nE, nRHS, nP, _ = strains.shape
    res = np.zeros_like(strains)
    for i in prange(nE):
        for j in prange(nRHS):
            for k in prange(nP):
                res[i, j, k] = D[i] @ strains[i, j, k]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def explode_kinetic_strains(kstrains: ndarray, nP: int):
    """
    ---
    (nE, nRHS, nP, nSTRE)
    """
    nE, nSTRE, nRHS = kstrains.shape
    res = np.zeros((nE, nRHS, nP, nSTRE), dtype=kstrains.dtype)
    for i in prange(nE):
        for j in prange(nRHS):
            for k in prange(nP):
                for m in prange(nSTRE):
                    res[i, j, k, m] = kstrains[i, m, j]
    return res


# -*- coding: utf-8 -*-
from enum import Enum, auto, unique

@unique
class DOF(Enum):
    UX = auto()
    UY = auto()
    UZ = auto()
    UYZ = auto()
    UXZ = auto()
    UXY = auto()


print(list(DOF))
print(list(map(lambda dof : dof.value, DOF)))


import numpy as np
from numpy import ndarray
from numba import njit, prange
__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_dcm(dcm: ndarray, N=2) -> ndarray:
    """
    Returns a direction cosine matrix for a node, 
    if the number of nodal degrees of freedom is a
    multiple of 3. 
    
    Parameters
    ----------
    dcm : numpy.ndarray
        A 3x3 direction cosine matrix for many elements.
    
    N : int, Optional
        The number of triplets that make up a nodal vector.
        Default is 2, which means 6 dofs per node.
        
    Returns
    -------
    numpy.ndarray
        A 3d numpy float array.
                 
    """
    nE = dcm.shape[0]
    res = np.zeros((nE, 3 * N, 3 * N), dtype=dcm.dtype)
    for i in prange(N):
        _i = i * 3
        i_ = _i + 3
        res[:, _i:i_, _i:i_] = dcm
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def element_dcm(nodal_dcm: ndarray, nNE: int = 2, nDOF: int=6):
    nE = nodal_dcm.shape[0]
    nEVAB = nNE * nDOF
    res = np.zeros((nE, nEVAB, nEVAB), dtype=nodal_dcm.dtype)
    for iNE in prange(nNE):
        i0, i1 = nDOF*iNE, nDOF * (iNE+1)
        for iE in prange(nE):
            res[iE, i0: i1, i0: i1] = nodal_dcm[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def element_transformation_matrices(Q: ndarray, nNE: int=2):
    nE = Q.shape[0]
    nEVAB = nNE * 6
    res = np.zeros((nE, nEVAB, nEVAB), dtype=Q.dtype)
    for iE in prange(nE):
        for j in prange(2*nNE):
            res[iE, 3*j : 3*(j+1), 3*j : 3*(j+1)] = Q[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_transformation_matrices(Q: ndarray):
    nE = Q.shape[0]
    res = np.zeros((nE, 6, 6), dtype=Q.dtype)
    for iE in prange(nE):
        for j in prange(2):
            res[iE, 3*j : 3*(j+1), 3*j : 3*(j+1)] = Q[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def transform_element_matrices_out(A: ndarray, Q: ndarray):
    res = np.zeros_like(A)
    for iE in prange(res.shape[0]):
        res[iE] = Q[iE].T @ A[iE] @ Q[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def transform_element_vectors_out(A: ndarray, Q: ndarray):
    """
    Transforms element load vectors from local to global.
    """
    res = np.zeros_like(A)
    for iE in prange(res.shape[0]):
        res[iE] = Q[iE].T @ A[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def transform_element_vectors_out_multi(A: ndarray, Q: ndarray):
    """
    Transforms multiple element load vectors from 
    local to global.
    """
    nE, nP = A.shape[:2]
    res = np.zeros_like(A)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP] = Q[iE].T @ A[iE, iP]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def transform_element_vectors_in(dofsol: ndarray, Q: ndarray):
    """
    Transforms element dof solutions from global to local.
    """
    res = np.zeros_like(dofsol)
    for i in prange(res.shape[0]):
        res[i] = Q[i] @ dofsol[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def transform_numint_forces_out(data: ndarray, dcm_G_L: ndarray):
    """
    Transforms internal forces of several elements
    and evaluation points from local frames
    to one global frame.
    
    Parameters
    ----------
    data: (nE, nP, nRHS, nDOFN) numpy.ndarray
        nE : number of elements
        nP : number of sampling points
        nRHS : number of datasets (eg. load cases)
        nDOFN : number of dofs of a node
        
    dcm_G_L : (nE, 3, 3) numpy array
        Array of DCM matrices from global to local.
    
    Returns
    -------
    (nE, nP, nRHS, nDOFN) numpy.ndarray
    """
    nE, nP, nC = data.shape[:3]
    res = np.zeros_like(data)
    for iE in prange(nE):
        for iP in prange(nP):
            for iC in prange(nC):
                res[iE, iP, iC, :] = dcm_G_L[iE].T @ data[iE, iP, iC, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def transform_numint_forces_in(data: ndarray, dcm_G_L: ndarray):
    """
    Transforms internal forces of several elements
    and numerical integration points form one global frame
    to several local frames.
    
    Parameters
    ----------
    data: (nE, nP, nRHS, nDOF) numpy.ndarray
        nE : number of elements
        nP : number of sampling points
        nRHS : number of datasets (eg. load cases)
        nDOFN : number of dofs of a node
        
    dcm_G_L : (nE, 3, 3) numpy array
        Array of DCM matrices from global to local.
    
    Returns
    -------
    (nE, nP, nRHS, nDOFN) numpy.ndarray   
    """
    nE, nP, nC = data.shape[:3]
    res = np.zeros_like(data)
    for i in prange(nE):
        for j in prange(nP):
            for k in prange(nC):
                res[i, j, k, :] = dcm_G_L[i] @ data[i, j, k, :]
    return res


# -*- coding: utf-8 -*-
import sympy as sy
from sympy import symbols, Matrix


def inv_sym_3x3(m: Matrix, as_adj_det=False):
    P11, P12, P13, P21, P22, P23, P31, P32, P33 = \
        symbols('P_{11} P_{12} P_{13} P_{21} P_{22} P_{23} P_{31} \
                P_{32} P_{33}', real=True)
    Pij = [[P11, P12, P13], [P21, P22, P23], [P31, P32, P33]]
    P = sy.Matrix(Pij)
    detP = P.det()
    adjP = P.adjugate()
    invP = adjP / detP
    subs = {s: r for s, r in zip(sy.flatten(P), sy.flatten(m))}
    if as_adj_det:
        return detP.subs(subs), adjP.subs(subs)
    else:
        return invP.subs(subs)



# -*- coding: utf-8 -*-
import numpy as np

from neumann.linalg import normalize as norm
from polymesh.utils import distances_of_points


def plot_path(ax, data, coords, *args, 
              normalize=False, markersize=3, 
              markercolor='r', title=None, **kwargs):
    x = np.cumsum(distances_of_points(coords))
    x = norm(x) if normalize else x
    ax.plot(x, data, zorder=1, color='k', lw=2)
    ax.scatter(x, data, s=markersize, zorder=2, color=markercolor)
    if title is not None:
        ax.set_title(title)
    return ax


# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np
from numpy import ndarray

from neumann.linalg.sparse.csr import csr_matrix as csr
from neumann.array import find1d, flatten2dC

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def fixity2d_to_dofs1d(fixity2d: np.ndarray, inds: np.ndarray = None):
    """
    Returns the indices of the degrees of freedoms
    being supressed. 

    Optionally, global indices of the rows in 'fixity2d' 
    array can be provided by the optional argument 'inds'.

    Parameters
    ----------
    fixity2d : np.ndarray(bool)[:, :]
        2d numpy array of booleans. It has as many rows as nodes, and as 
        many columns as derees of freedom per node in the model. A True
        value means that the corresponding dof is fully supressed.

    inds : np.ndarray, optional
        1d numpy array of integers, default is None. If provided, it should
        list the global indices of the nodes, for which the data is provided
        by the array 'fixity2d'.

    Returns
    -------
    dofinds : np.ndarray
        1d numpy array of integers. 

    """
    ndof = fixity2d.shape[1]
    args = np.argwhere(fixity2d == True)
    N = args.shape[0]
    res = np.zeros(N, dtype=args.dtype)
    for i in prange(N):
        res[i] = args[i, 0]*ndof + args[i, 1]
    if inds is None:
        return res
    else:
        dofinds = np.zeros_like(res)
        for i in prange(len(inds)):
            for j in prange(ndof):
                dofinds[i*ndof + j] = inds[i]*ndof + j
        return dofinds[res]


@njit(nogil=True, parallel=True, cache=__cache)
def nodes2d_to_dofs1d(inds: np.ndarray, values: np.ndarray):
    """
    Returns a tuple of degree of freedom indices and data, 
    based on a nodal definition.

    Parameters
    ----------
    inds : np.ndarray
        1d numpy array of integers, listing global node indices.

    values : int of shape (nN, nDOF, nRHS)
        3d numpy array of floats, listing values for each node
        in 'inds'.

    Returns
    -------
    dofinds : np.ndarray
        1d numpy array of integers, denoting global dof indices.

    dofvals : np.ndarray of shape (nN * nDOF, nRHS)
        2d numpy array of floats, denoting values on dofs in 'dofinds'.

    """
    nN, dof_per_node, nRHS = values.shape
    dofinds = np.zeros((nN * dof_per_node), dtype=inds.dtype)
    dofvals = np.zeros((nN * dof_per_node, nRHS), dtype=values.dtype)
    for node in prange(nN):
        for dof in prange(dof_per_node):
            i = node * dof_per_node + dof
            dofinds[i] = inds[node] * dof_per_node + dof
            for rhs in prange(nRHS):
                dofvals[i, rhs] = values[node, dof, rhs]
    return dofinds, dofvals


@njit(nogil=True, cache=__cache, parallel=True)
def weighted_stiffness_bulk(K: np.ndarray, weights: np.ndarray):
    """
    Returns a weighted stiffness matrix.

    Parameters
    ----------
    K : np.ndarray
        2d numpy array of floats

    weights : np.ndarray
        1d numpy array of floats

    Returns
    -------
    Kw : np.ndarray
        2d numpy array of floats

    Notes
    -----
    (1) It is assumed that the first axis of K runs
        along the elements.

    """
    nE = len(weights)
    res = np.zeros_like(K)
    for i in prange(nE):
        res[i, :] = weights[i] * K[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def irows_icols_bulk(edofs: np.ndarray):
    """
    Returns row and column index data for several finite elements.

    Parameters
    ----------
    edofs : np.ndarray
        2d numpy array. Each row has the meaning of global degree of 
        freedom numbering for a given finite element.

    Returns
    -------
    irows, icols : np.ndarray, np.ndarray
        Global indices of the rows and columns of the stiffness matrices
        of the elements.

    Notes
    -----
    The implementation assumes that every cell has the same number of
    degrees of freedom.

    """
    nE, nNE = edofs.shape
    nTOTV = nNE * nNE
    irows = np.zeros((nE, nTOTV), dtype=edofs.dtype)
    icols = np.zeros((nE, nTOTV), dtype=edofs.dtype)
    for iNE in prange(nNE):
        for jNE in prange(nNE):
            i = iNE * nNE + jNE
            for iE in prange(nE):
                irows[iE, i] = edofs[iE, iNE]
                icols[iE, i] = edofs[iE, jNE]
    return irows, icols


@njit(nogil=True, cache=__cache, parallel=True)
def irows_icols_bulk_filtered(edofs: np.ndarray, inds: np.ndarray):
    """
    Returns row and column index data for finite elements specified
    by the index array `inds`.

    Parameters
    ----------
    edofs : np.ndarray
        2d numpy array. Each row has the meaning of global degree of 
        freedom numbering for a given finite element.

    inds: np.ndarray
        1d numpy array of integers specifying active elements in an assembly.

    Returns
    -------
    irows, icols : np.ndarray, np.ndarray
        Global indices of the rows and columns of the stiffness matrices
        of the elements.

    Notes
    -----
    The implementation assumes that every cell has the same number of
    degrees of freedom.

    """
    nI = len(inds)
    nNE = edofs.shape[1]
    nTOTV = nNE * nNE
    irows = np.zeros((nI, nTOTV), dtype=edofs.dtype)
    icols = np.zeros((nI, nTOTV), dtype=edofs.dtype)
    for iNE in prange(nNE):
        for jNE in prange(nNE):
            i = iNE * nNE + jNE
            for iI in prange(nI):
                irows[iI, i] = edofs[inds[iI], iNE]
                icols[iI, i] = edofs[inds[iI], jNE]
    return irows, icols


@njit(nogil=True, cache=__cache, parallel=True)
def topo_to_gnum(topo: np.ndarray, ndofn: int):
    """
    Returns global dof numbering based on element 
    topology data.

    Parameters
    ----------
    topo : np.ndarray
        2d numpy array of integers. Topology array listing global
        node numbers for several elements.

    ndofn : int
        Number of degrees of freedoms per node.

    Returns
    -------
    gnum : np.ndarray
        2d numpy array of integers.

    """
    nE, nNE = topo.shape
    gnum = np.zeros((nE, nNE*ndofn), dtype=topo.dtype)
    for i in prange(nE):
        for j in prange(nNE):
            for k in prange(ndofn):
                gnum[i, j*ndofn + k] = topo[i, j]*ndofn + k
    return gnum


@njit(nogil=True, cache=__cache, parallel=True)
def topo_to_gnum_jagged(topo, gnum, ndofn: int):
    """
    Returns global dof numbering based on element 
    topology data.

    Parameters
    ----------
    topo : np.ndarray
        2d array of integers. Topology array listing global
        node numbers for several elements.
        
    gnum : np.ndarray
        2d buffer array for the results.

    ndofn : int
        Number of degrees of freedoms per node.

    Returns
    -------
    gnum : np.ndarray
        2d numpy array of integers.

    """
    nE, _ = topo.shape
    for i in prange(nE):
        for j in prange(len(topo[i])):
            for k in prange(ndofn):
                gnum[i, j*ndofn + k] = topo[i, j]*ndofn + k
    return gnum


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def assemble_load_vector(values: ndarray, gnum: ndarray, N: int = -1):
    """
    Returns global dof numbering based on element 
    topology data.

    Parameters
    ----------
    values : np.ndarray of shape (nE, nEVAB, nRHS)
        3d numpy array of floats, representing element data.
        The length of the second axis matches the the number of
        degrees of freedom per cell.

    gnum : int
        Global indices of local degrees of freedoms of elements.

    N : int, Optional.
        The number of total unknowns in the system. Must be specified correcly,
        to get a vector the same size of the global system. If not specified, it is
        inherited from 'gnum' (as 'gnum.max() + 1'), but this can lead to a chopped 
        array.

    Returns
    -------
    np.ndarray
        2d numpy array of integers with a shape of (N, nRHS), where nRHS is the number
        if right hand sizes (load cases).

    """
    nE, nEVAB, nRHS = values.shape 
    if N < 0:
        N = gnum.max() + 1
    res = np.zeros((N, nRHS), dtype=values.dtype)
    for i in range(nE):
        for j in range(nEVAB):
            for k in prange(nRHS):
                res[gnum[i, j], k] += values[i, j, k]
    return res


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def penalty_factor_matrix(cellfixity: ndarray, shp: ndarray):
    nE, nNE, nDOF = cellfixity.shape
    N = nDOF * nNE
    res = np.zeros((nE, N, N), dtype=shp.dtype)
    for iE in prange(nE):
        for iNE in prange(nNE):
            for iDOF in prange(nDOF):
                fix = cellfixity[iE, iNE, iDOF]
                for jNE in prange(nNE):
                    i = jNE * nDOF + iDOF
                    for kNE in prange(nNE):
                        j = kNE * nDOF + iDOF
                        res[iE, i, j] += fix * shp[iNE, jNE] * shp[iNE, kNE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def approximation_matrix(ndf: ndarray, NDOFN: int):
    """Returns a matrix of approximation coefficients 
    for all elements."""
    nE, nNE = ndf.shape[:2]
    N = nNE * NDOFN
    nappr = np.eye(N, dtype=ndf.dtype)
    res = np.zeros((nE, N, N), dtype=ndf.dtype)
    for iE in prange(nE):
        for i in prange(nNE):
            for j in prange(nNE):
                for ii in prange(NDOFN):
                    for jj in prange(NDOFN):
                        res[iE, i*2 + ii, j*2 + jj] = nappr[i, j] * ndf[iE, i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_approximation_matrix(ndf: ndarray):
    """Returns a matrix of nodal approximation coefficients 
    for all elements."""
    nE, nNE = ndf.shape[:2]
    nappr = np.eye(nNE, dtype=ndf.dtype)
    res = np.zeros((nE, nNE, nNE), dtype=ndf.dtype)
    for iE in prange(nE):
        for i in prange(nNE):
            for j in prange(nNE):
                res[iE, i, j] = nappr[i, j] * ndf[iE, i]
    return res


@njit(nogil=True, cache=__cache)
def nodal_compatibility_factors(nam_csr_tot: csr, nam: ndarray, topo: ndarray):
    indptr = nam_csr_tot.indptr
    indices = nam_csr_tot.indices
    data = nam_csr_tot.data
    nN = len(indptr)-1
    widths = np.zeros(nN, dtype=indptr.dtype)
    for iN in range(nN):
        widths[iN] = indptr[iN+1] - indptr[iN]
    factors = dict()
    nreg = dict()
    for iN in range(nN):
        factors[iN] = np.zeros((widths[iN], widths[iN]), dtype=nam.dtype)
        nreg[iN] = indices[indptr[iN]: indptr[iN+1]]
    nE, nNE = topo.shape
    for iE in range(nE):
        topoE = topo[iE]
        for jNE in range(nNE):
            nID = topo[iE, jNE]
            dataN = data[indptr[nID]: indptr[nID+1]]
            topoN = nreg[nID]
            dataNE = np.zeros(widths[nID], dtype=nam.dtype)
            imap = find1d(topoE, topoN)
            dataNE[imap] = nam[iE, jNE]
            rNE = dataN - dataNE  # residual
            factors[nID] += np.outer(rNE, rNE)
    return factors, nreg


@njit(nogil=True, parallel=True, cache=__cache)
def compatibility_factors_to_coo(ncf: dict, nreg: dict):
    """
    ncf : nodal_compatibility_factors
    """
    nN = len(ncf)
    widths = np.zeros(nN, dtype=np.int32)
    for iN in prange(nN):
        widths[iN] = len(nreg[iN])
    shapes = (widths**2).astype(np.int64)
    N = np.sum(shapes)
    data = np.zeros(N, dtype=np.float64)
    rows = np.zeros(N, dtype=np.int32)
    cols = np.zeros(N, dtype=np.int32)
    c = 0
    for iN in range(nN):
        data[c: c + shapes[iN]] = flatten2dC(ncf[iN])
        nNN = widths[iN]
        for jNN in prange(nNN):
            for kNN in prange(nNN):
                rows[c + jNN*nNN + kNN] = nreg[iN][jNN]
                cols[c + jNN*nNN + kNN] = nreg[iN][kNN]
        c += shapes[iN]
    return data, rows, cols


@njit(nogil=True, parallel=True, cache=__cache)
def topo1d_to_gnum1d(topo1d: ndarray, NDOFN: int):
    nN = topo1d.shape[0]
    gnum1d = np.zeros(nN * NDOFN, dtype=topo1d.dtype)
    for i in prange(nN):
        for j in prange(NDOFN):
            gnum1d[i*NDOFN + j] = topo1d[i] * NDOFN + j
    return gnum1d


@njit(nogil=True, parallel=True, cache=__cache)
def ncf_to_cf(ncf: ndarray, NDOFN: int):
    nN = ncf.shape[0]
    cf = np.zeros((nN * NDOFN, nN * NDOFN), dtype=ncf.dtype)
    for i in prange(nN):
        for ii in prange(NDOFN):
            for j in prange(nN):
                cf[i*NDOFN + ii, j*NDOFN + ii] = ncf[i, j]
    return cf


@njit(nogil=True, cache=__cache)
def compatibility_factors(ncf: dict, nreg: dict, NDOFN: int):
    """ncf : nodal_compatibility_factors"""
    nN = len(ncf)
    widths = np.zeros(nN, dtype=np.int32)
    for iN in prange(nN):
        widths[iN] = len(nreg[iN])
    cf = dict()
    reg = dict()
    for iN in range(nN):
        cf[iN] = ncf_to_cf(ncf[iN], NDOFN)
        reg[iN] = topo1d_to_gnum1d(nreg[iN], NDOFN)
    return cf, reg


@njit(nogil=True, parallel=True, cache=__cache)
def tr_cells_1d_in(A: ndarray, Q: ndarray):
    """
    Transforms element vectors (like the load vector) from global to local.
    
    Parameters
    ----------
    A : 3d NumPy float array of shape (nE, nEVAB)
        Array of coefficients to transform.
        
    Q : 3d NumPy float array of shape (nE, nEVAB, nEVAB)
        Transformation matrices.
        
    Returns
    -------
    numpy array
        NumPy array with the same shape as 'A'
    """
    res = np.zeros_like(A)
    for iE in prange(res.shape[0]):
        res[iE] = Q[iE] @ A[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tr_cells_1d_out(A: ndarray, Q: ndarray):
    """
    Transforms element vectors (like the load vector) from local to global.
    (nE, nNE * nDOF)
    """
    res = np.zeros_like(A)
    for iE in prange(res.shape[0]):
        res[iE] = Q[iE].T @ A[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tr_cells_1d_in_multi(A: ndarray, Q: ndarray):
    """
    Transforms element vectors (like the load vector) from local to global
    for multiple cases.
    (nE, nRHS, nNE * nDOF)
    """
    res = np.zeros_like(A)
    for iE in prange(res.shape[0]):
        for jRHS in prange(res.shape[1]):
            res[iE, jRHS] = Q[iE] @ A[iE, jRHS]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tr_cells_1d_out_multi(A: ndarray, Q: ndarray):
    """
    Transforms element vectors (like the load vector) from local to global
    for multiple cases.
    A (nE, nRHS, nP * nDOF)
    Q (nE, nP * nDOF)
    """
    res = np.zeros_like(A)
    for iE in prange(res.shape[0]):
        for jRHS in prange(res.shape[1]):
            res[iE, jRHS] = Q[iE].T @ A[iE, jRHS]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def element_dof_solution_bulk(dofsol1d: ndarray, gnum: ndarray):
    """
    dofsol (nN * nDOF, nRHS)
    gnum (nE, nEVAB)
    ---
    (nE, nEVAB, nRHS)
    """
    nRHS = dofsol1d.shape[1]
    nE, nEVAB = gnum.shape
    res = np.zeros((nE, nEVAB, nRHS), dtype=dofsol1d.dtype)
    for i in prange(nE):
        for j in prange(nRHS):
            res[i, :, j] = dofsol1d[gnum[i, :], j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def transform_stiffness(K: ndarray, dcm: ndarray):
    """
    Transforms element stiffness matrices from local to global.
    """
    res = np.zeros_like(K)
    for iE in prange(res.shape[0]):
        res[iE] = dcm[iE].T @ K[iE] @ dcm[iE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def element_dofmap_bulk(dofmap: ndarray, nDOF: int, nNODE: int):
    nDOF_ = dofmap.shape[0]
    res = np.zeros(nNODE * nDOF_, dtype=dofmap.dtype)
    for i in prange(nNODE):
        for j in prange(nDOF_):
            res[i*nDOF_ + j] = i*nDOF + dofmap[j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def expand_stiffness_bulk(K_in: ndarray, K_out: ndarray, dofmap: ndarray):
    """
    Expands the local stiffness matrix into a standard form.
    """
    nDOF = dofmap.shape[0]    
    for i in prange(nDOF):
        for j in prange(nDOF):
            K_out[:, dofmap[i], dofmap[j]] = K_in[:, i, j]
    return K_out


@njit(nogil=True, parallel=True, cache=__cache)
def pull_submatrix(A, r, c):
    nR = r.shape[0]
    nC = c.shape[0]
    res = np.zeros((nR, nC), dtype=A.dtype)
    for i in prange(nR):
        for j in prange(nC):
            res[i, j] = A[r[i], c[j]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def push_submatrix(A, Asub, r, c, new=True):
    nR = r.shape[0]
    nC = c.shape[0]
    res = np.zeros_like(A)
    if not new:
        res[:, :] = A
    for i in prange(nR):
        for j in prange(nC):
            res[r[i], c[j]] = Asub[i, j]
    return res
    

@njit(nogil=True, parallel=True, cache=__cache)
def constrain_local_stiffness_bulk(K: ndarray, factors: ndarray):
    """
    Returns the condensed stiffness matrices representing constraints
    on the internal forces of the elements (eg. hinges).
        
    Currently this solution is only able to handle two states, being totally free 
    and being fully constrained. The factors are expected to be numbers between
    0 and 1, where dofs with a factor > 0.5 are assumed to be the constrained ones.
    
    Note
    ----
    numba-jitted in nopython mode
    
    Parameters
    ----------
    K : numpy.ndarray
        3d float array, the stiffness matrices of several elements of the same kind.
        
    factors: numpy.ndarray
        2d float array of connectivity facotors for each dof of every element.
                    
    Note
    ----
    This solution applies the idea of static condensation.
    
    Returns
    -------
    numpy.ndarray
        The constrained stiffness matrices with the same shape as `K`.
        
    """
    nE, _, _ = K.shape
    res = np.zeros_like(K)
    for iE in prange(nE):
        b = np.where(factors[iE] > 0.5)[0]
        i = np.where(factors[iE] <= 0.5)[0]
        Kbb = pull_submatrix(K[iE], b, b)
        Kii = pull_submatrix(K[iE], i, i)
        Kib = pull_submatrix(K[iE], i, b)
        Kbi = pull_submatrix(K[iE], b, i)
        Kbb -= Kbi @ np.linalg.inv(Kii) @ Kib
        res[iE] = push_submatrix(K[iE], Kbb, b, b, True)
    return res


def assert_min_stiffness_bulk(K: ndarray, minval = 1e-12):
    inds = np.arange(K.shape[-1])
    d = K[:, inds, inds]
    eid, vid = np.where(d < minval)
    K[eid, vid, vid] = minval
    return K


@njit(nogil=True, parallel=True, cache=__cache)
def internal_forces(K: ndarray, dofsol: ndarray):
    """
    Transforms element stiffness matrices from local to global.
    ---
    (nE, nRHS, nEVAB)
    """
    nE, nRHS, nEVAB  = dofsol.shape
    res = np.zeros_like(dofsol)
    for i in prange(nE):
        for j in prange(nRHS):
            res[i, j] = K[i] @ dofsol[i, j]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def nodal_mass_matrix_data(nodal_masses: ndarray, ndof:int=6):
    N = nodal_masses.shape[0]
    res = np.zeros((N*ndof))
    for i in prange(N):
        res[i*ndof : i*ndof + 3] = nodal_masses[i]
    return res



# -*- coding: utf-8 -*-
from typing import Union, Tuple, Callable

import numpy as np
from scipy.sparse import coo_matrix

from linkeddeepdict import LinkedDeepDict

from dewloosh.core.wrapping import Wrapper

from neumann import squeeze
from neumann.array import repeat

from .mesh import FemMesh
from .femsolver import FemSolver


__all__ = ['Structure']


def flatten_pd(default=True):
    def decorator(fnc: Callable):
        def inner(*args, **kwargs):
            if kwargs.get('flatten', default):
                x = fnc(*args, **kwargs)
                if len(x.shape) == 2:
                    return x.flatten()
                else:
                    nN, nDOFN, nRHS = x.shape
                    return x.reshape((nN * nDOFN, nRHS))
            else:
                return fnc(*args, **kwargs)
        inner.__doc__ = fnc.__doc__
        return inner
    return decorator


class Structure(Wrapper):
    """
    A higher level class to manage solid structures.

    Parameters
    ----------
    mesh : :class:`FemMesh`
        A finite element mesh.

    """

    def __init__(self, *args, mesh: FemMesh = None, constraints=None, **kwargs):
        if not isinstance(mesh, FemMesh):
            raise NotImplementedError
        super().__init__(wrap=mesh)
        assert mesh is not None, "Some kind of a finite element mesh must be \
            provided with keyword 'mesh'!"
        self.summary = LinkedDeepDict()
        self.solver = 'scipy'
        self.Solver = None
        self._constraints = constraints

    @property
    def mesh(self) -> FemMesh:
        """
        Returns the underlying mesh object.
        """
        return self._wrapped

    @mesh.setter
    def mesh(self, value: FemMesh):
        """
        Sets the underlying mesh object.
        """
        self._wrapped = value

    @property
    def Solver(self) -> FemSolver:
        """
        Returns the solver object, which is initialized in the preprocessing stage.

        See also
        --------
        :func:`preprocess`

        Returns
        -------
        :class:`.femsolver.FemSolver`

        """
        return self._solver

    @Solver.setter
    def Solver(self, value: Solver):
        self._solver = value

    def initialize(self, *args, **kwargs) -> 'Structure':
        """
        Initializes the structure. This includes data initialization 
        for the cells and happens during the preprocessor. Returns the object 
        for continuation.

        See also
        --------
        :func:`preprocess`

        """
        self.summary = LinkedDeepDict()
        blocks = self.mesh.cellblocks(inclusive=True)
        for block in blocks:
            nE = len(block.celldata)

            # populate material stiffness matrices
            if not 'mat' in block.celldata.fields:
                C = block.material_stiffness_matrix()
                if not len(C.shape) == 3:
                    C = repeat(block.material_stiffness_matrix(), nE)
                block.celldata._wrapped['mat'] = C

            # populate frames
            if not 'frames' in block.celldata.fields:
                frames = repeat(block.frame.show(), nE)
                block.celldata._wrapped['frames'] = frames
        return self

    def preprocess(self, *args, **kwargs) -> 'Structure':
        """
        Initializes the database and converts the discription into standard form.
        Returns the object for continuation.

        See also
        --------
        :func:`initialize`
        :func:`to_standard_form`

        """
        self.initialize(*args, **kwargs)
        # self.mesh.nodal_distribution_factors(store=True, key='ndf')  # sets mesh.celldata.ndf
        self.Solver = self.to_standard_form(*args, **kwargs)
        return self

    def process(self, *args, summary=False, **kwargs) -> 'Structure':
        """
        Performs a linear elastostatic solution and returns the structure
        and returns the object for continuation.

        Parameters
        ----------
        summary : bool, Optional
            Appends related data to the summary. Default is False.

        """
        self.Solver.linsolve(*args, summary=True, **kwargs)
        if summary:
            self.summary['linsolve'] = self.Solver.summary[-1]
        return self

    def to_standard_form(self, *args, ensure_comp=False, solver=FemSolver, **kwargs) -> Solver:
        """
        Returns a solver of the problem in standard form. Creation of the solver happens
        during the preprocessing stage.
        """
        mesh = self._wrapped
        f = mesh.load_vector()
        Kp_coo = mesh.penalty_matrix_coo(ensure_comp=ensure_comp, **kwargs)
        K_bulk = mesh.elastic_stiffness_matrix(*args, sparse=False, **kwargs)
        gnum = mesh.element_dof_numbering()
        solvertype = solver if solver is not None else FemSolver
        return solvertype(K_bulk, Kp_coo, f, gnum, regular=False)

    def linsolve(self, *args, summary=False, postproc=True, **kwargs) -> 'Structure':
        """
        Performs a linear elastostatic calculation with pre- and 
        post-processing.

        Parameters
        ----------
        summary : bool, Optional
            Appends to the overall summary if True. This is available
            as `obj.summary`. Default is False.

        postproc : bool, Optional
            To do postprocessing or not. Default is True.

        """
        self.preprocess(*args, summary=summary, **kwargs)
        self.process(*args, summary=summary, **kwargs)
        if postproc:
            return self.postprocess(*args, summary=summary, **kwargs)
        else:
            return self

    def natural_circular_frequencies(self, *args, **kwargs) -> Tuple:
        """
        Returns natural circular frequencies. The call forwards all parameters
        to :func:`sigmaepslion.solid.fem.dyn.natural_circular_frequencies`, 
        see the docs there for the details. 

        Parameters
        ----------
        return_vectors : bool, Optional
            To return eigenvectors or not. Default is False.

        See also
        --------
        :func:`sigmaepslion.solid.fem.dyn.natural_circular_frequencies`

        Returns
        -------
        :class:`numpy.ndarray`
            The natural circular frequencies.

        :class:`numpy.ndarray`
            The eigenvectors, only if 'return_vectors' is True.

        """
        if self.Solver.M is not None:
            self.consistent_mass_matrix()
        return self.Solver.natural_circular_frequencies(*args, **kwargs)

    def elastic_stiffness_matrix(self, *args, **kwargs) -> Union[np.ndarray, coo_matrix]:
        """
        Returns the elastic stiffness matrix of the structure in dense or sparse format.

        Returns
        -------
        :class:`numpy.ndarray` or :class:`scipy.sparse.coo_matrix`

        """
        return self.mesh.elastic_stiffness_matrix(*args, **kwargs)

    def penalty_stiffness_matrix(self, *args, **kwargs) -> coo_matrix:
        """
        Returns the penalty stiffness matrix of the structure.

        Returns
        -------
        :class:`scipy.sparse.coo_matrix`

        """
        return self.mesh.penalty_matrix_coo(*args, **kwargs)

    def consistent_mass_matrix(self, *args, **kwargs) -> Union[np.ndarray, coo_matrix]:
        """
        Returns the consistent mass matrix of the structure.

        Notes
        -----
        If there are nodal masses defined, only sparse output is 
        available at the moment.

        Returns
        -------
        :class:`numpy.ndarray` or :class:`scipy.sparse.coo_matrix`

        """
        M = self.mesh.consistent_mass_matrix(*args, **kwargs)
        if self.Solver is not None:
            if self.Solver.M is None:
                self.Solver.M = M
        return M

    @flatten_pd(False)
    @squeeze(True)
    def nodal_dof_solution(self, *args, flatten: bool = False, squeeze: bool = True,
                           store: str = None, **kwargs) -> np.ndarray:
        """
        Returns the vector of nodal displacements and optionally stores the result
        with a specified key.

        Parameters
        ----------
        flatten : bool, Optional
            If True, the result is a 1d array, otherwise 2d. Default is False.

        squeeze : bool, Optional
            Calls :func:`numpy.squeeze` on the result. This might be relevant for 
            single element operations, or if there is only one load case, etc. Then,
            it depends on the enviroment if a standard shape is desirable to maintain or not.
            Default is True.

        store : str, Optional
            If a string is provided, the resulting data can be later accessed as an attribute
            of the pointdata of the mesh object. Default is None.

        """
        mesh = self._wrapped
        nDOFN = mesh.NDOFN
        nN = mesh.number_of_points()

        # store dof solution
        u = self.Solver.u
        nRHS = 1 if len(u.shape) == 1 else u.shape[-1]
        if nRHS == 1:
            res = np.reshape(u, (nN, nDOFN))
        else:
            res = np.reshape(u, (nN, nDOFN, nRHS))

        if isinstance(store, str):
            mesh.pointdata[store] = res

        return res

    @flatten_pd(False)
    @squeeze(True)
    def reaction_forces(self, *args, flatten: bool = False, squeeze: bool = True,
                        store: str = None, **kwargs) -> np.ndarray:
        """
        Returns the vector of reaction forces.

        Parameters
        ----------
        flatten : bool, Optional
            If True, the result is a 1d array, otherwise 2d. Default is False.

        squeeze : bool, Optional
            Calls :func:`numpy.squeeze` on the result. This might be relevant for 
            single element operations, or if there is only one load case, etc. Then,
            it depends on the enviroment if a standard shape is desirable to maintain or not.
            Default is True.

        store : str, Optional
            If a string is provided, the resulting data can be later accessed as an attribute
            of the pointdata of the mesh object. Default is None.

        """
        mesh = self._wrapped
        nDOFN = mesh.NDOFN
        nN = mesh.number_of_points()

        # store dof solution
        r = self.Solver.r
        nRHS = 1 if len(r.shape) == 1 else r.shape[-1]
        if nRHS == 1:
            res = np.reshape(r, (nN, nDOFN))
        else:
            res = np.reshape(r, (nN, nDOFN, nRHS))

        if isinstance(store, str):
            mesh.pointdata[store] = res

        return res

    @flatten_pd(False)
    @squeeze(True)
    def nodal_forces(self, *args, flatten: bool = False, squeeze: bool = True,
                     store: str = None, **kwargs) -> np.ndarray:
        """
        Returns the vector of nodal forces.

        Parameters
        ----------
        flatten : bool, Optional
            If True, the result is a 1d array, otherwise 2d. Default is False.

        squeeze : bool, Optional
            Calls :func:`numpy.squeeze` on the result. This might be relevant for 
            single element operations, or if there is only one load case, etc. Then,
            it depends on the enviroment if a standard shape is desirable to maintain or not.
            Default is True.

        store : str, Optional
            If a string is provided, the resulting data can be later accessed as an attribute
            of the pointdata of the mesh object. Default is None.

        """

        mesh = self._wrapped
        nDOFN = mesh.NDOFN
        nN = mesh.number_of_points()

        # store nodal loads
        f = self.Solver.f
        nRHS = 1 if len(f.shape) == 1 else f.shape[-1]
        if nRHS == 1:
            res = np.reshape(f, (nN, nDOFN))
        else:
            res = np.reshape(f, (nN, nDOFN, nRHS))

        if isinstance(store, str):
            mesh.pointdata[store] = res

        return res

    @squeeze(True)
    def internal_forces(self, *args, flatten=False, squeeze=True, **kwargs) -> np.ndarray:
        """
        Returns the internal forces for one or more elements.
        """
        return self.mesh.internal_forces(*args, flatten=flatten, squeeze=False, **kwargs)

    def postprocess(self, *args, cleanup=False, **kwargs) -> 'Structure':
        """
        Runs general postprocessing of the solution data. Returns the object for continuation.

        The calculations produce the following point-related data, stored in the pointdata
        object of the root object of the finite element mesh:

            - nodal dof solution, available with key 'dofsol'

            - vector of nodal forces, available with key 'forces'

            - vector of reaction forces, available with key 'reactions'

        For information on the generated cell-related data see :func:`sigmaepsilon.solid.fem.mesh.FemMesh.postprocess`.

        Parameters
        ----------
        *args : tuple, Optional
            Optional positional arguments to the postprocessing engine of the mesh.
            Only if postproc is True.

        **kwargs : dict, Optional
            Optional keyword arguments to the postprocessing engine of the mesh.
            Only if postproc is True.

        cleanup : bool, Optional
            If True, data related to the calculation is freed up and the model
            needs to be reinitialized to run calculations again. Default is False.

        See also
        --------
        :class:`sigmaepsilon.solid.fem.mesh.FemMesh`

        Note
        ----
        This can be triggered automatically during linear solution when calling :func:`linsolve`
        with `postproc=True`.

        Returns
        -------
        :class:`Structure`

        """
        mesh = self._wrapped
        mesh.postprocess(*args, **kwargs)

        # clean up
        _ = self.cleanup() if cleanup else None

        return self

    def cleanup(self) -> 'Structure':
        """
        Destroys the solver and returns the object for continuation.

        Returns
        -------
        :class:`Structure`

        """
        self.Solver = None
        return self



# -*- coding: utf-8 -*-
import numpy as np
#from dewloosh.solid.model.mindlin.utils import \
#    stiffness_data_Mindlin, pproc_Mindlin_3D
from numpy import sin, cos, ndarray as nparray, pi as PI
from numba import njit, prange

from neumann.array import atleast2d, atleast3d, atleast4d, itype_of_ftype


__all__ = ['postproc']


UZ, ROTX, ROTY, CX, CY, CXY, EXZ, EYZ, MX, MY, MXY, QX, QY = list(range(13))


def postproc(ABDS: nparray, points: nparray, *args,
             size: tuple = None, shape: tuple = None, squeeze=True,
             angles=None, bounds=None, shear_factors=None,
             C_126=None, C_45=None, separate=True, res2d=None,
             solution=None, model: str = 'mindlin', loads: nparray, **kwargs):
    nD = points.shape[1]
    ABDS = atleast3d(ABDS)
    dtype = ABDS.dtype
    itype = itype_of_ftype(ABDS.dtype)

    # 2d postproc
    if res2d is None and solution is not None:
        assert size is not None
        assert shape is not None
        assert points is not None
        if model.lower() in ['mindlin', 'm']:
            res2d = pproc_Mindlin_2d(np.array(size).astype(dtype),
                                     np.array(shape).astype(itype),
                                     atleast2d(points)[:, :2].astype(dtype),
                                     atleast4d(solution).astype(dtype),
                                     ABDS[:, :3, :3], ABDS[:, 3:, 3:])
        elif model.lower() in ['kirchhoff', 'k']:
            res2d = pproc_Kirchhoff_2d(np.array(size).astype(dtype),
                                       np.array(shape).astype(itype),
                                       atleast2d(points)[:, :2].astype(dtype),
                                       atleast3d(solution).astype(dtype),
                                       ABDS[:, :3, :3], ABDS[:, 3:, 3:], loads)
        res2d = np.sum(res2d, axis=0)

    # 3d postproc
    assert res2d is not None
    assert points is not None
    assert points.shape[0] == res2d.shape[2]
    if nD == 2 and C_126 is not None:
        # lamination scheme may br provided
        assert bounds is not None
        raise NotImplementedError
    elif nD == 3 and C_126 is not None:
        """
        res3d = pproc_Mindlin_Navier_3D(ABDS, C_126, C_45, bounds, points,
                                        res2d, shear_factors=shear_factors,
                                        squeeze=squeeze, angles=angles,
                                        separate=separate)
        return res3d if not squeeze else np.squeeze(res3d)
        """
        raise NotImplementedError
    else:
        return res2d if not squeeze else np.squeeze(res2d)


@njit(nogil=True, parallel=True, cache=True)
def pproc_Mindlin_2d(size, shape: nparray, points: nparray,
                     solution: nparray, D: nparray, S: nparray):
    """
    JIT-compiled function that calculates post-processing quantities 
    at selected ponts for multiple left- and right-hand sides.       

    Parameters
    ----------
    size : tuple 
        (Lx, Ly) : size of domain

    shape : tuple 
        (M, N) : number of harmonic terms involved in x and y
                 directions

    points : numpy.ndarray[nP, 2] 
        2d array of point coordinates

    solution : numpy.ndarray[nRHS, M * N, 3] 
        results of a Navier solution as a 3d array

    D : numpy.ndarray[nLHS, 3, 3] 
        3d array of bending stiffness terms

    S : numpy.ndarray[nLHS, 2, 2]
        3d array of shear stiffness terms

    Returns
    -------
    numpy.ndarray[M * N, nRHS, nLHS, nP, ...] 
        numpy array of post-processing items. The indices along 
        the last axpis denote the following quantities:

            0 : displacement z
            1 : rotation x
            2 : rotation y
            3 : curvature x
            4 : curvature y
            5 : curvature xy
            6 : shear strain xz
            7 : shear strain yz
            8 : moment y
            9 : moment y
            10 : moment xy
            11 : shear force x
            12 : shear force y
    """
    Lx, Ly = size
    M, N = shape
    nP = points.shape[0]
    nRHS, nLHS = solution.shape[:2]
    res2d = np.zeros((M * N, nRHS, nLHS, nP, 13), dtype=D.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], \
                D[iLHS, 1, 1], D[iLHS, 2, 2]
            S44, S55 = S[iLHS, 0, 0], S[iLHS, 1, 1]
            for iP in prange(nP):
                xp, yp = points[iP, :2]
                for m in prange(1, M + 1):
                    Sm = sin(PI*m*xp/Lx)
                    Cm = cos(PI*m*xp/Lx)
                    for n in prange(1, N + 1):
                        iMN = (m-1) * N + n - 1
                        Amn, Bmn, Cmn = solution[iRHS, iLHS, iMN]
                        Sn = sin(PI*n*yp/Ly)
                        Cn = cos(PI*n*yp/Ly)
                        res2d[iMN, iRHS, iLHS, iP, UZ] = Cmn * Sm*Sn
                        res2d[iMN, iRHS, iLHS, iP, ROTX] = Amn * Sm*Cn
                        res2d[iMN, iRHS, iLHS, iP, ROTY] = Bmn * Sn*Cm
                        res2d[iMN, iRHS, iLHS, iP, CX] = -PI*Bmn * m*Sm*Sn/Lx
                        res2d[iMN, iRHS, iLHS, iP, CY] = PI*Amn * n*Sm*Sn/Ly
                        res2d[iMN, iRHS, iLHS, iP, CXY] = -PI*Amn*m*Cm*Cn/Lx + \
                            PI*Bmn*n*Cm*Cn/Ly
                        res2d[iMN, iRHS, iLHS, iP, EXZ] = Bmn*Sn*Cm + \
                            PI*Cmn*m*Sn*Cm/Lx
                        res2d[iMN, iRHS, iLHS, iP, EYZ] = -Amn*Sm*Cn + \
                            PI*Cmn*n*Sm*Cn/Ly
                        res2d[iMN, iRHS, iLHS, iP, MX] = PI*Amn*D12*n*Sm*Sn/Ly - \
                            PI*Bmn*D11*m*Sm*Sn/Lx
                        res2d[iMN, iRHS, iLHS, iP, MY] = PI*Amn*D22*n*Sm*Sn/Ly - \
                            PI*Bmn*D12*m*Sm*Sn/Lx
                        res2d[iMN, iRHS, iLHS, iP, MXY] = -PI*Amn*D66*m*Cm*Cn/Lx + \
                            PI*Bmn*D66*n*Cm*Cn/Ly
                        res2d[iMN, iRHS, iLHS, iP, QX] = Bmn*S55*Sn*Cm + \
                            PI*Cmn*S55*m*Sn*Cm/Lx
                        res2d[iMN, iRHS, iLHS, iP, QY] = -Amn*S44*Sm*Cn + \
                            PI*Cmn*S44*n*Sm*Cn/Ly
    return res2d


@njit(nogil=True, parallel=True, cache=True)
def pproc_Kirchhoff_2d(size, shape: nparray, points: nparray,
                       solution: nparray, D: nparray, S: nparray,
                       loads: nparray):
    """
    JIT-compiled function that calculates post-processing quantities 
    at selected ponts for multiple left- and right-hand sides.       

    Parameters
    ----------
    size : tuple 
        (Lx, Ly) : size of domain

    shape : tuple 
        (M, N) : number of harmonic terms involved in x and y
                 directions

    points : numpy.ndarray[nP, 2] 
        2d array of point coordinates

    solution : numpy.ndarray[nRHS, M * N] 
        results of a Navier solution as a 2d array

    D : numpy.ndarray[nLHS, 3, 3] 
        3d array of bending stiffness terms

    S : numpy.ndarray[nLHS, 2, 2]
        3d array of shear stiffness terms

    Returns
    -------
    numpy.ndarray[M * N, nRHS, nLHS, nP, ...] 
        numpy array of post-processing items. The indices along 
        the last axpis denote the following quantities:

            0 : displacement z
            1 : rotation x
            2 : rotation y
            3 : curvature x
            4 : curvature y
            5 : curvature xy
            6 : shear strain xz
            7 : shear strain yz
            8 : moment y
            9 : moment y
            10 : moment xy
            11 : shear force x
            12 : shear force y
    """
    Lx, Ly = size
    M, N = shape
    nP = points.shape[0]
    nRHS, nLHS = solution.shape[:2]
    res2d = np.zeros((M * N, nRHS, nLHS, nP, 13), dtype=D.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], \
                D[iLHS, 1, 1], D[iLHS, 2, 2]
            S44, S55 = S[iLHS, 0, 0], S[iLHS, 1, 1]
            for iP in prange(nP):
                x, y = points[iP, :2]
                for m in prange(1, M + 1):
                    Sm = sin(PI*m*x/Lx)
                    Cm = cos(PI*m*x/Lx)
                    for n in prange(1, N + 1):
                        Sn = sin(PI*n*y/Ly)
                        Cn = cos(PI*n*y/Ly)
                        iMN = (m-1) * N + n - 1
                        Cmn = solution[iRHS, iLHS, iMN]
                        qxx, qyy = loads[iRHS, iMN, :2]
                        res2d[iMN, iRHS, iLHS, iP, UZ] = Cmn * Sm*Sn
                        res2d[iMN, iRHS, iLHS, iP, ROTX] = PI * Cmn*n*Sm*Cn/Ly
                        res2d[iMN, iRHS, iLHS, iP, ROTY] = -PI * Cmn*m*Sn*Cm/Lx
                        res2d[iMN, iRHS, iLHS, iP, CX] = PI**2 * \
                            Cmn * m**2*Sm*Sn/Lx**2
                        res2d[iMN, iRHS, iLHS, iP, CY] = PI**2 * \
                            Cmn * n**2*Sm*Sn/Ly**2
                        res2d[iMN, iRHS, iLHS, iP, CXY] = - \
                            2*PI**2 * Cmn*m*n*Cm*Cn/(Lx*Ly)
                        res2d[iMN, iRHS, iLHS, iP, EXZ] = PI**3*Cmn*D11*m**3*Sn*Cm/(Lx**3*S44) + \
                            PI**3*Cmn*D12*m*n**2*Sn*Cm/(Lx*Ly**2*S44) + \
                            2*PI**3*Cmn*D66*m*n**2*Sn*Cm / \
                            (Lx*Ly**2*S44) + qxx*Sn*Cm/S44
                        res2d[iMN, iRHS, iLHS, iP, EYZ] = PI**3*Cmn*D12*m**2*n*Sm*Cn/(Lx**2*Ly*S55) + \
                            PI**3*Cmn*D22*n**3*Sm*Cn/(Ly**3*S55) + \
                            2*PI**3*Cmn*D66*m**2*n*Sm*Cn / \
                            (Lx**2*Ly*S55) + qyy*Sm*Cn/S55
                        res2d[iMN, iRHS, iLHS, iP, MX] = PI**2*Cmn*D11*m**2*Sm*Sn/Lx**2 + \
                            PI**2*Cmn*D12*n**2 * Sm*Sn/Ly**2
                        res2d[iMN, iRHS, iLHS, iP, MY] = PI**2*Cmn*D12*m**2*Sm*Sn/Lx**2 + \
                            PI**2*Cmn*D22*n**2 * Sm*Sn/Ly**2
                        res2d[iMN, iRHS, iLHS, iP, MXY] = - \
                            2*PI**2*Cmn * D66*m*n*Cm*Cn/(Lx*Ly)
                        res2d[iMN, iRHS, iLHS, iP, QX] = PI**3*Cmn*D11*m**3*Sn*Cm/Lx**3 + \
                            PI**3*Cmn*D12*m*n**2*Sn*Cm/(Lx*Ly**2) + \
                            2*PI**3*Cmn*D66*m*n**2*Sn*Cm/(Lx*Ly**2) + qxx*Sn*Cm
                        res2d[iMN, iRHS, iLHS, iP, QY] = PI**3*Cmn*D12*m**2*n*Sm*Cn/(Lx**2*Ly) + \
                            PI**3*Cmn*D22*n**3*Sm*Cn/Ly**3 + \
                            2*PI**3*Cmn*D66*m**2*n*Sm*Cn/(Lx**2*Ly) + qyy*Sm*Cn
    return res2d

"""
def pproc_Mindlin_Navier_3D(ABDS: nparray,
                            C_126: nparray, C_45: nparray,
                            bounds: nparray, points: nparray,
                            res2d: nparray, *args,
                            angles=None, separate=True,
                            squeeze=False, shear_factors: nparray = None,
                            **kwargs):
    nLHS, nRHS, nP, _ = res2d.shape
    strains2d = np.zeros((nLHS, nRHS, nP, 8), dtype=ABDS.dtype)
    strains2d[:, :, :, 3:8] = res2d[:, :, :, 3:8]
    ABDS = atleast3d(ABDS)
    res3d = pproc_Mindlin_3D(ABDS, C_126, C_45, bounds, points, strains2d,
                             separate=separate,
                             squeeze=squeeze, angles=angles,
                             shear_factors=shear_factors)
    return res3d


def shell_stiffness_data(shell):
    shell.stiffness_matrixp()
    layers = shell.layers()
    C_126, C_45 = [], []
    for layer in layers:
        Cm = layer.material.stiffness_matrixp()
        C_126.append(Cm[0:3, 0:3])
        C_45.append(Cm[3:, 3:])
    C_126 = np.stack(C_126)
    C_45 = np.stack(C_45)
    angles = np.stack([layer.angle for layer in layers])
    bounds = np.stack([[layer.tmin, layer.tmaxp]
                       for layer in layers])
    ABDS, shear_corrections, shear_factors = \
        stiffness_data_Mindlin(C_126, C_45, angles, bounds)
    ABDS[-2:, -2:] *= shear_corrections
    return C_126, C_45, bounds, angles, ABDS, shear_factors
"""

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
from ..config import __hassympy__, __haspylatex__

if __haspylatex__:
    from pylatex import NoEscape, Package


def append_preamble(doc):
    r"""
    Tools related to displaying math. It's a bit like the numpy of latex, must
    have stuff.
    """
    doc.packages.append(Package('amsmath'))

    r"""
    Sympy uses the 'operatorname' command frequently to print symbols.
    """
    doc.packages.append(Package('amsopn'))

    r"""
    To automatically break long equations into multiple lines.
    """
    doc.packages.append(Package('breqn'))

    r"""
    mathtools provides us with the \coloneqq command, for defining equality
    symbol ':='
    """
    doc.packages.append(Package('mathtools'))

    r"""
    Misc
    """
    doc.packages.append(Package('enumitem'))  # to customize enumerations
    doc.packages.append(Package('xcolor'))  # colors
    doc.packages.append(Package('lmodern'))  # high quality fonts
    
    return doc


def expr_to_ltx(lhs, rhs, *args, env='{equation}', sign='=',
                dfrac=False, pre=None, post=None, **kwargs):
    if dfrac:
        lhs = lhs.replace('frac', 'dfrac')
        rhs = rhs.replace('frac', 'dfrac')
    if isinstance(pre, str):
        lhs = ' '.join([pre, lhs])
    if isinstance(post, str):
        rhs = ' '.join([rhs, post])
    return NoEscape(
        r"""
        \begin{env}
            {lhs} {sign} {rhs}
        \end{env}
        """.format(env = env,
                   lhs = lhs,
                   sign = sign,
                   rhs = rhs
                   )
        )


def expr_to_ltx_breqn(lhs, rhs, *args, env='{dmath}', **kwargs):
    return expr_to_ltx(lhs, rhs, *args, env=env, **kwargs)





# -*- coding: utf-8 -*-
import json
import numpy as np
from numpy import sin, cos, ndarray, pi as PI
from numba import njit, prange
from collections import Iterable

from linkeddeepdict import LinkedDeepDict
from linkeddeepdict.tools.dtk import parsedicts_addr
from dewloosh.core.tools import allinkwargs, popfromkwargs, float_to_str_sig

from neumann import squeeze
from neumann.array import atleast1d, atleast2d, atleast3d


class LoadGroup(LinkedDeepDict):
    _typestr_ = None

    def __init__(self, *args, Navier=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._Navier = Navier

    def Navier(self):
        return self.root()._Navier

    def blocks(self, *args, inclusive=False, blocktype=None,
               deep=True, **kwargs):
        dtype = LoadGroup if blocktype is None else blocktype
        return self.containers(self, inclusive=inclusive,
                               dtype=dtype, deep=deep)

    def load_cases(self, *args, inclusive=True, **kwargs):
        return filter(lambda i: i.__class__._typestr_ is not None, 
                      self.blocks(*args, inclusive=inclusive, **kwargs))

    @staticmethod
    def string_to_dtype(string: str = None):
        if string == 'group':
            return LoadGroup
        elif string == 'rect':
            return RectLoad
        elif string == 'point':
            return PointLoad
        else:
            raise NotImplementedError

    def dump(self, path, *args, mode='w', indent=4, **kwargs):
        with open(path, mode) as f:
            json.dump(self.to_dict(), f, indent=indent)

    @classmethod
    def from_json(cls, path: str = None, **kwargs):
        if path is not None:
            with open(path, 'r') as f:
                d = json.load(f)
            return cls.from_dict(d, **kwargs)

    def encode(self, *args, **kwargs) -> dict:
        """
        Overwrite this in child implementations.
        """
        res = {}
        cls = type(self)
        res = {
            'type': cls._typestr_,
            'key': self.key,
        }
        return res

    @classmethod
    def decode(cls, d: dict = None, *args, **kwargs):
        """
        Overwrite this in child implementations.
        """
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            'key': d.pop('key', None),
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    def to_dict(self):
        res = self.encode()
        for key, value in self.items():
            if isinstance(value, LoadGroup):
                res[key] = value.to_dict()
            else:
                res[key] = value
        return res

    @staticmethod
    def from_dict(d: dict = None, **kwargs) -> 'LoadGroup':
        if 'type' in d:
            cls = LoadGroup.string_to_dtype(d['type'])
        else:  
            cls = LoadGroup
        res = cls(**d)
        for addr, value in parsedicts_addr(d, inclusive=True):
            if len(addr) == 0:
                continue
            if 'type' in value:
                cls = LoadGroup.string_to_dtype(value['type'])
            else: 
                cls = LoadGroup
            value['key'] = addr[-1]
            res[addr] = cls(**value)
        return res

    def __repr__(self):
        return 'LoadGroup(%s)' % (dict.__repr__(self))


class RectLoad(LoadGroup):
    _typestr_ = 'rect'

    def __init__(self, *args, value=None, **kwargs):
        self.points = RectLoad.get_coords(kwargs)
        self.value = np.array(value, dtype=float)
        super().__init__(*args, **kwargs)

    def encode(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            'type': cls._typestr_,
            'key': self.key,
            'region': float_to_str_sig(self.region(), sig=6),
            'value': float_to_str_sig(self.value, sig=6),
        }
        return res

    @classmethod
    def decode(cls, d: dict = None, *args, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        points = RectLoad.get_coords(d)
        clskwargs = {
            'key': d.pop('key', None),
            'points': points,
            'value': np.array(d.pop('value'), dtype=float)
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    @staticmethod
    def get_coords(d: dict = None, *args, **kwargs):
        points = None
        if d is None:
            d = kwargs
        try:
            if 'points' in d:
                points = np.array(d.pop('points'))
            elif 'region' in d:
                x0, y0, w, h = np.array(d.pop('region'))
                points = np.array([[x0, y0], [x0 + w, y0 + h]])
            elif allinkwargs(['xy', 'w', 'h'], **d):
                (x0, y0), w, h = popfromkwargs(['xy', 'w', 'h'], d)
                points = np.array([[x0, y0], [x0 + w, y0 + h]])
            elif allinkwargs(['center', 'w', 'h'], **d):
                (xc, yc), w, h = popfromkwargs(['center', 'w', 'h'], d)
                points = np.array([[xc - w/2, yc - h/2],
                                   [xc + w/2, yc + h/2]])
        except Exception as e:
            print(e)
            return None
        return points

    def region(self):
        if self.points is not None:
            return points_to_region(self.points)
        return None, None, None, None

    def rhs(self, *args, **kwargs):
        Navier = kwargs.get('Navier', self.Navier())
        return rhs_rect_const(Navier.size, Navier.shape, self.value, self.points)

    def __repr__(self):
        return 'RectLoad(%s)' % (dict.__repr__(self))


@squeeze(True)
def rhs_rect_const(size: tuple, shape: tuple, values: np.ndarray,
                   points: np.ndarray, *args, **kwargs):
    """
    Returns coefficients for continous loads in the order [mx, my, fz].
        mx : bending moment aroung global x
        my : bending moment aroung global y
        fz : force along global z
    """
    return _rect_const_(size, shape, atleast2d(values), atleast3d(points))


@njit(nogil=True, cache=True)
def __rect_const__(size: tuple, m: int, n: int, xc: float, yc: float,
                   w: float, h: float, values: ndarray):
    Lx, Ly = size
    mx, my, fz = values
    return np.array([16*mx*sin((1/2)*PI*m*w/Lx) *
                     sin((1/2)*PI*h*n/Ly)*sin(PI*n*yc/Ly) *
                     cos(PI*m*xc/Lx)/(PI**2*m*n),
                     16*my*sin((1/2)*PI*m*w/Lx) *
                     sin(PI*m*xc/Lx)*sin((1/2)*PI*h*n/Ly) *
                     cos(PI*n*yc/Ly)/(PI**2*m*n),
                     16*fz*sin((1/2)*PI*m*w/Lx)*sin(PI*m*xc/Lx) *
                     sin((1/2)*PI*h*n/Ly)*sin(PI*n*yc/Ly)/(PI**2*m*n)])


@njit(nogil=True, parallel=True, cache=True)
def _rect_const_(size: tuple, shape: tuple, values: np.ndarray,
                points: np.ndarray):
    nR = values.shape[0]
    M, N = shape
    rhs = np.zeros((nR, M * N, 3), dtype=points.dtype)
    for iR in prange(nR):
        xmin, ymin = points[iR, 0]
        xmax, ymax = points[iR, 1]
        xc = (xmin + xmax)/2
        yc = (ymin + ymax)/2
        w = np.abs(xmax - xmin)
        h = np.abs(ymax - ymin)
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                mn = (m - 1) * N + n - 1
                rhs[iR, mn, :] = \
                    __rect_const__(size, m, n, xc, yc, w, h, values[iR])
    return rhs


class LineLoad(LoadGroup):
    _typestr_ = 'line'

    def __init__(self, *args, points=None, values=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._points = np.array(points, dtype=float)
        self._values = np.array(values, dtype=float)
        
    @classmethod
    def decode(cls, d: dict = None, *args, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            'key': d.pop('key', None),
            'points': np.array(d.pop('points')),
            'values': np.array(d.pop('values')),
        }
        clskwargs.update(d)
        return cls(**clskwargs)
    
    def encode(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            'type': cls._typestr_,
            'key': self.key,
            'points': float_to_str_sig(self._points, sig=6),
            'values': float_to_str_sig(self._values, sig=6),
        }
        return res
    
    def rhs(self, *args, **kwargs):
        raise NotImplementedError
        Navier = kwargs.get('Navier', self.Navier())
        return rhs_conc(Navier.size, Navier.shape, self.value, self.point)

    def __repr__(self):
        return 'LineLoad(%s)' % (dict.__repr__(self))


class PointLoad(LoadGroup):
    _typestr_ = 'point'

    def __init__(self, *args, point=None, value=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.point = np.array(point, dtype=float)
        self.value = np.array(value, dtype=float)

    @classmethod
    def decode(cls, d: dict = None, *args, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            'key': d.pop('key', None),
            'point': np.array(d.pop('point')),
            'value': np.array(d.pop('value')),
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    def encode(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            'type': cls._typestr_,
            'key': self.key,
            'point': float_to_str_sig(self.point, sig=6),
            'value': float_to_str_sig(self.value, sig=6),
        }
        return res

    def rhs(self, *args, **kwargs):
        Navier = kwargs.get('Navier', self.Navier())
        return rhs_conc_2d(Navier.size, Navier.shape, self.value, self.point)

    def __repr__(self):
        return 'PointLoad(%s)' % (dict.__repr__(self))


@squeeze(True)
def rhs_conc_2d(size: tuple, shape: tuple, values: np.ndarray,
                points: np.ndarray, *args, **kwargs):
    if isinstance(size, Iterable):
        return _conc2d_(size, shape, atleast2d(values), atleast2d(points))
    else:
        raise NotImplementedError
        return _conc1d_(size, shape, atleast2d(values), atleast1d(points))


@njit(nogil=True, parallel=True, cache=True)
def _conc1d_(size: tuple, shape: tuple, values: ndarray, points: ndarray):
    # FIXME Not verified
    nRHS = values.shape[0]
    Lx = size
    c = 2 / Lx
    N = shape
    rhs = np.zeros((nRHS, N, 2), dtype=points.dtype)
    PI = np.pi
    for iRHS in prange(nRHS):
        x = points[iRHS]
        my, fz = values[iRHS]
        Sx = PI * x / Lx
        for i in prange(N):
            rhs[iRHS, i, :] = c
            rhs[iRHS, i, 0] *= my * cos(i * Sx)
            rhs[iRHS, i, 1] *= fz * sin(i * Sx)
    return rhs


@njit(nogil=True, parallel=True, cache=True)
def _conc2d_(size: tuple, shape: tuple, values: ndarray, points: ndarray):
    nRHS = values.shape[0]
    Lx, Ly = size
    c = 4 / Lx / Ly
    M, N = shape
    rhs = np.zeros((nRHS, M * N, 3), dtype=points.dtype)
    PI = np.pi
    for iRHS in prange(nRHS):
        x, y = points[iRHS]
        mx, my, fz = values[iRHS]
        Sx = PI * x / Lx
        Sy = PI * y / Ly
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                mn = (m - 1) * N + n - 1
                rhs[iRHS, mn, :] = c
                rhs[iRHS, mn, 0] *= mx * cos(m * Sx) * sin(n * Sy)
                rhs[iRHS, mn, 1] *= my * sin(m * Sx) * cos(n * Sy)
                rhs[iRHS, mn, 2] *= fz * sin(m * Sx) * sin(n * Sy)
    return rhs


def points_to_region(points: ndarray):
    xmin = points[:, 0].min()
    ymin = points[:, 1].min()
    xmax = points[:, 0].max()
    ymax = points[:, 1].max()
    return xmin, ymin, xmax - xmin, ymax - ymin



# -*- coding: utf-8 -*-
import numpy as np
from numba import njit, prange
from numpy import ndarray, pi as PI

from neumann import squeeze
from neumann.array import atleast3d


@squeeze(True)
def lhs_Navier(size: tuple, shape: tuple, *args, D: ndarray,
               S: ndarray = None, model: str = 'mindlin', **kwargs):
    """
    Returns coefficient matrix for a Navier solution, for a single or 
    multiple left-hand sides.

    Parameters
    ----------
    size : tuple
        Tuple of floats, containing the sizes of the rectagle.

    shape : tuple
        Tuple of integers, containing the number of harmonic terms
        included in both directions.

    D : numpy.ndarray
        2d or 3d float array of bending stiffnesses.

    S : numpy.ndarray, Optional
        2d or 3d float array of shear stiffnesses. Default is None.

    squeeze : boolean, optional
        Removes single-dimensional entries from the shape of the 
        resulting array. Default is True.

    Returns
    -------
    numpy.ndarray
        3d or 4d float array of coefficients. The shape depends on
        the shape of the input parameters.
        
    """
    if model.lower() in ['mindlin', 'm']:
        return lhs_Mindlin(size, shape, atleast3d(D), atleast3d(S))
    elif model.lower() in ['kirchhoff', 'k']:
        return lhs_Kirchhoff(size, shape, atleast3d(D))


@njit(nogil=True, parallel=True, cache=True)
def lhs_Mindlin(size: tuple, shape: tuple, D: np.ndarray, S: np.ndarray):
    """
    JIT compiled function, that returns coefficient matrices for a Navier 
    solution for multiple left-hand sides.

    Parameters
    ----------
    size : tuple
        Tuple of floats, containing the sizes of the rectagle.

    shape : tuple
        Tuple of integers, containing the number of harmonic terms
        included in both directions.

    D : numpy.ndarray
        3d float array of bending stiffnesses.

    S : numpy.ndarray
        3d float array of shear stiffnesses.

    Returns
    -------
    numpy.ndarray
        4d float array of coefficients.
        
    """
    Lx, Ly = size
    nLHS = D.shape[0]
    M, N = shape
    res = np.zeros((nLHS, M * N, 3, 3), dtype=D.dtype)
    for iLHS in prange(nLHS):
        D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], \
            D[iLHS, 1, 1], D[iLHS, 2, 2]
        S44, S55 = S[iLHS, 0, 0], S[iLHS, 1, 1]
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                iMN = (m - 1) * N + n - 1
                res[iLHS, iMN, 0, 0] = -PI**2*D22*n**2 / \
                    Ly**2 - PI**2*D66*m**2/Lx**2 - S44
                res[iLHS, iMN, 0, 1] = PI**2*D12*m * \
                    n/(Lx*Ly) + PI**2*D66*m*n/(Lx*Ly)
                res[iLHS, iMN, 0, 2] = PI*S44*n/Ly
                res[iLHS, iMN, 1, 0] = -PI**2*D12*m * \
                    n/(Lx*Ly) - PI**2*D66*m*n/(Lx*Ly)
                res[iLHS, iMN, 1, 1] = PI**2*D11*m**2 / \
                    Lx**2 + PI**2*D66*n**2/Ly**2 + S55
                res[iLHS, iMN, 1, 2] = PI*S55*m/Lx
                res[iLHS, iMN, 2, 0] = -PI*S44*n/Ly
                res[iLHS, iMN, 2, 1] = PI*S55*m/Lx
                res[iLHS, iMN, 2, 2] = PI**2*S44 * \
                    n**2/Ly**2 + PI**2*S55*m**2/Lx**2
    return res


@njit(nogil=True, parallel=True, cache=True)
def lhs_Kirchhoff(size: tuple, shape: tuple, D: np.ndarray):
    """
    JIT compiled function, that returns coefficient matrices for a Navier 
    solution for multiple left-hand sides.

    Parameters
    ----------
    size : tuple
        Tuple of floats, containing the sizes of the rectagle.

    shape : tuple
        Tuple of integers, containing the number of harmonic terms
        included in both directions.

    D : numpy.ndarray
        3d float array of bending stiffnesses.

    Returns
    -------
    numpy.ndarray
        2d float array of coefficients.
        
    """
    Lx, Ly = size
    nLHS = D.shape[0]
    M, N = shape
    res = np.zeros((nLHS, M * N), dtype=D.dtype)
    for iLHS in prange(nLHS):
        D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], \
            D[iLHS, 1, 1], D[iLHS, 2, 2]
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                iMN = (m - 1) * N + n - 1
                res[iLHS, iMN] = PI**4*D11*m**4/Lx**4 + \
                    2*PI**4*D12*m**2*n**2/(Lx**2*Ly**2) + \
                        PI**4*D22*n**4/Ly**4 + \
                            4*PI**4*D66*m**2*n**2/(Lx**2*Ly**2)
    return res



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange

from neumann import squeeze
from neumann.linalg.solve import npsolve


@squeeze(True)
def linsolve(LHS: ndarray, RHS: ndarray, *args, **kwarg):
    if len(LHS.shape) > 2:
        return linsolve_Mindlin(LHS, RHS)
    else:
        return linsolve_Kirchhoff(LHS, RHS)


@njit(nogil=True, parallel=True, cache=True)
def linsolve_Kirchhoff(A: ndarray, B: ndarray):
    nLHS, nMN = A.shape
    nRHS = B.shape[0]
    res = np.zeros((nRHS, nLHS, nMN))
    for i in prange(nRHS):
        for j in prange(nLHS):
            for k in prange(nMN):
                res[i, j, k] = B[i, k] / A[j, k]
    return res


@njit(nogil=True, parallel=True, cache=True)
def linsolve_Mindlin(A: ndarray, B: ndarray):
    nLHS, nMN = A.shape[:2]
    nRHS = B.shape[0]
    res = np.zeros((nRHS, nLHS, nMN, 3))
    for i in prange(nRHS):
        for j in prange(nLHS):
            for k in prange(nMN):
                #res[i, j, k] = inv3x3(A[j, k]) @ B[i, k]
                res[i, j, k] = npsolve(A[j, k], B[i, k]) 
    return res


# -*- coding: utf-8 -*-
from copy import deepcopy
from functools import partial

import numpy as np
from randomdict import RandomDict

from linkeddeepdict import LinkedDeepDict
from polymesh.space import frames_of_lines

from ..model.bernoulli.section import BeamSection, get_section
from .cells import B2, B3


__all__ = ['_get_bernoulli_metadata_']


def _dofdict2list(d: dict, dofs: list, default=0.0) -> list:
    missing_keys = filter(lambda k: k not in d, dofs)
    d.update({k: default for k in missing_keys})
    return [d[k] for k in dofs]


def _get_nodal_dof_vector(*args, **kwargs):
    if isinstance(args[0], list):
        return args[0]
    if isinstance(args[0], dict):
        return _dofdict2list(*args, **kwargs)
    raise NotImplementedError


def _get_bernoulli_metadata_(data, *args, dofs=None, return_imap=False):

    if dofs is None:
        dofs = ('UX', 'UY', 'UZ', 'ROTX', 'ROTY', 'ROTZ')

    get_V6 = partial(_get_nodal_dof_vector, dofs=dofs)

    d = LinkedDeepDict(deepcopy(data))

    imap = LinkedDeepDict()
    # This is going to be filled up with inverse maps of indices
    # and keys. This allows for arbitrary keys in the input.
    # You might need this to translate between input data and the output.

    # ---------------------------------------------------------------
    # --------------------------- POINTS ----------------------------
    # ---------------------------------------------------------------

    # allocate arrays for point-like data
    pkey = 'points' if 'points' in d else 'nodes'
    assert pkey in d
    # allocate data related to points
    # these objects are going to be filled up in the next for loop
    nP = len(d[pkey])
    coords = np.zeros((nP, 3), dtype=float)  # coordinates
    fixity = np.zeros((nP, 6), dtype=bool)  # essential bc
    mass = np.zeros((nP,), dtype=float)  # nodal masses
    loads = LinkedDeepDict()
    # collect data and create inverse mapping
    i_p = 0
    id_to_key = {}
    for key, p in data[pkey].items():
        id_to_key[i_p] = key
        if isinstance(p, list):
            # only coordinates
            coords[i_p] = p
            d[pkey] = {
                "x": p,
                "id": i_p,
            }
        elif isinstance(p, dict):
            # several data
            d[pkey]['id'] = i_p
            coords[i_p] = p['x']
            if 'mass' in p:
                mass[i_p] = p['mass']
            if 'load' in p:
                for case, value in p['load'].items():
                    loads[case, 'conc', i_p] = get_V6(value, default=0.0)
            if 'fixity' in p:
                fixity[i_p] = get_V6(p['fixity'], default=False)
        i_p += 1
    imap['points'] = deepcopy(id_to_key)
    id_to_key = {}

    # ---------------------------------------------------------------
    # --------------- FRAMES, MATERIALS AND SECTIONS ----------------
    # ---------------------------------------------------------------

    # get materials and add some data to the records
    if 'frames' in data:
        i_f = 0
        id_to_key = {}
        for framekey in data['frames'].keys():
            id_to_key[i_f] = framekey
            i_f += 1
        imap['frames'] = deepcopy(id_to_key)
        id_to_key = {}

    # get materials and add some data to the records
    if 'materials' in data:
        i_m = 0
        id_to_key = {}
        for materialkey in data['materials'].keys():
            id_to_key[i_m] = materialkey
            d['materials'][materialkey]['cells'] = []  # we fill this up later
            i_m += 1
        imap['materials'] = deepcopy(id_to_key)
        id_to_key = {}

    # get sections and add some data to the records
    if 'sections' in data:
        i_s = 0
        id_to_key = {}
        for sectionkey, section in data['sections'].items():
            if 'type' in section:
                s = get_section(section['type'], **section)
                if s is not None:
                    bs = BeamSection(wrap=s)
                    bs.calculate_section_properties()
                    props = bs.get_section_properties()
                    props['obj'] = bs  # store section object
                else:
                    raise NotImplementedError("Section type not supported.")
            else:
                # raw input, at least keys A, Ix, Iy, Iz must be present
                props = {'obj': None}
                props.update(section)
            if 'material' in section:
                props.pdate(data['materials'][section['material']])
            props['id'] = i_s  # add an index
            props['cells'] = []  # we fill this up later
            d['sections'][sectionkey].update(**props)
            id_to_key[i_s] = sectionkey
            i_s += 1
        imap['sections'] = deepcopy(id_to_key)
        id_to_key = {}

    # ---------------------------------------------------------------
    # ---------------------------- CELLS ----------------------------
    # ---------------------------------------------------------------

    # FIXME This assumes a reguar input at the moment

    cells = RandomDict(**data['cells'])
    nE = len(cells)
    random_cell = cells.random_value()
    tkey = 'nodes' if 'nodes' in random_cell else 'topology'
    nNE = len(random_cell[tkey])

    topo = np.zeros((nE, nNE), dtype=int)  # topology
    conn = np.zeros((nE, nNE, 6), dtype=int)  # connectivity
    Hooke = np.zeros((nE, 4, 4), dtype=float)  # material model
    frames = np.zeros((nE, 3, 3), dtype=float)  # coordinate frames
    i_c = 0
    id_to_key = {}
    for cellkey, cell in cells.items():
        id_to_key[i_c] = key
        # topology
        topo[i_c] = cell[tkey]
        # connectivity
        if 'connectivity' in data['cells'][cellkey]:
            conn[i_c] = d['cells'][cellkey]['connectivity']
        # material
        if 'material' in cell:
            mid = cell['material']
        else:
            # material is defined for the section
            assert 'section' in cell, "A section must be defined"
            assert 'material' in cell['section'], "A material must be defined for the section."
            mid = data['sections'][cell['section']]['material']
            m = data['materials'][mid]
        try:
            m = data['materials'][mid]
        except KeyError:
            if 'materials' in data:
                raise KeyError("Material Id {} npt found.".format(mid))
            else:
                raise KeyError("Materials are not defined.")
        # section
        if 'section' in cell:
            s = d['sections'][cell['section']]
            d['sections'][cell['section']]['cells'].append(i_c)
            d['materials'][mid]['cells'].append(i_c)
        # build Hook's model
        E, nu = m['E'], m['nu']
        A, Ix, Iy, Iz = s['A'], s['Ix'], s['Iy'], s['Iz']
        G = E / (2 * (1 + nu))
        Hooke[i_c] = np.array([
            [E*A, 0, 0, 0],
            [0, G*Ix, 0, 0],
            [0, 0, E*Iy, 0],
            [0, 0, 0, E*Iz]
        ])

        # coordinate frame
        if 'frame' in cell:
            if isinstance(cell['frame'], dict):
                frame = np.array(
                    [
                        cell['frame']['i'],
                        cell['frame']['j'],
                        cell['frame']['k']
                    ]
                )
            elif isinstance(cell['frame'], list):
                frame = np.array(cell['frame'])
            else:
                frame = data['frames'][cell['frame']]
            frames[i_c] = frame

        # body loads
        if 'load' in cell:
            # nDOF values per element
            for case, value in cell['load'].items():
                loads[case, 'body', i_c] = get_V6(value, default=0.0)

        # temperature load
        if 'heat' in cell:
            # 1 value per element
            for case, value in cell['heat'].items():
                loads[case, 'heat', i_c] = value

        # register indices
        """d['cells'][cellkey]['id'] = i_c
        d['cells'][cellkey]['material_id'] = m['id']
        d['cells'][cellkey]['section_id'] = s['id']"""
        i_c += 1
    imap['cells'] = deepcopy(id_to_key)
    id_to_key = {}
    # create coordinate frames
    # topology must be known for this implementation
    # FIXME This could be integrated into the block above, by
    # creating a single evaluation version of 'frames_of_lines'
    # -> 'frame_of_line'
    frames_auto = frames_of_lines(coords, topo)
    i_c = 0
    for cellkey, cell in cells.items():
        if not 'frame' in cell:
            frames[i_c] = frames_auto[i_c]
        i_c += 1

    # ---------------------------------------------------------------
    # ---------------------------- LOADS ----------------------------
    # ---------------------------------------------------------------
    # assembly of the load vectors
    nLC = len(loads)

    # vector of nodal loads
    nodal_loads = np.zeros((nP, 6, nLC))

    # body load vector - for every dof for every node for element
    body_loads = np.zeros((nE, nNE, 6, nLC), dtype=float)  # vector of

    # strain load vector - for every strain component for every element
    strain_loads = np.zeros((nE, 4, nLC), dtype=float)
    i_lc = 0
    cid_to_ckey = imap['cells']
    for loadkey, lc in loads.items():
        if 'conc' in lc:
            for i, values in lc['conc'].items():
                nodal_loads[i, :, i_lc] = values
        if "body" in lc:
            for i, values in lc['body'].items():
                for iNE in range(nNE):
                    body_loads[i, iNE, :, i_lc] = values
        if "heat" in lc:
            for cid, heat in lc['heat'].items():
                ckey = cid_to_ckey[cid]
                mkey = data['cells'][ckey]['material']
                coeff = data['materials'][mkey]['thermal-coefficient']
                strain_loads[int(cellkey), 0, i_lc] = heat * coeff
        d['loads'][loadkey]['id'] = i_lc
        i_lc += 1

    d['imap'] = imap
    return d, dict(coords=coords, topo=topo, loads=nodal_loads, fixity=fixity,
                   model=Hooke, body_loads=body_loads, strain_loads=strain_loads,
                   frames=frames, mass=mass)



# -*- coding: utf-8 -*-
from typing import Union, List

from scipy.sparse import coo_matrix
import numpy as np

from neumann import squeeze
from neumann.array import atleast3d

from polymesh import PolyData

from .pointdata import PointData
from .cells.celldata import CellData
from .preproc import fem_load_vector, fem_penalty_matrix_coo, fem_nodal_mass_matrix_coo
from .cells import TET4, TET10, H8, H27


class FemMesh(PolyData):
    """
    A descendant of :class:`polymesh.PolyData` to handle polygonal meshes for the **Finite Element Method**.

    Parameters
    ----------
    pd : :class:`polymesh.PolyData` or :class:`polymesh.CellData`, Optional
        A PolyData or a CellData instance. Dafault is None.

    cd : :class:`polymesh.CellData`, Optional
        A CellData instance, if the first argument is provided. Dafault is None.

    coords : :class:`numpy.ndarray`, Optional
        2d numpy array of floats, describing a pointcloud. Default is None.

    topo : :class:`numpy.ndarray`, Optional
        2d numpy array of integers, describing the topology of a polygonal mesh. 
        Default is None.

    celltype : int, Optional
        An integer spcifying a valid celltype.

    Note
    ----
    The suggested way of usage is to create the pointdata and celldata
    objects in advance, and provide them as the firts argument when building the model.
    See below for how it works!

    Examples
    --------
    >>> from sigmaepsilon.solid import FemMesh
    >>> from polymesh.grid import grid
    >>> size = Lx, Ly, Lz = 100, 100, 100
    >>> shape = nx, ny, nz = 10, 10, 10
    >>> coords, topo = grid(size=size, shape=shape, eshape='H27')
    >>> pd = FemMesh(coords=coords)
    >>> pd['A']['Part1'] = FemMesh(topo=topo[:10])
    >>> pd['B']['Part2'] = FemMesh(topo=topo[10:-10])
    >>> pd['C']['Part3'] = FemMesh(topo=topo[-10:])

    See also
    --------
    :class:`polymesh.PolyData`
    :class:`polymesh.CellData`

    """

    dofs = ('UX', 'UY', 'UZ', 'ROTX', 'ROTY', 'ROTZ')
    NDOFN = 6
    _point_class_ = PointData

    def __init__(self, *args, model=None, fixity=None, loads=None, body_loads=None,
                 strain_loads=None, t=None, density=None, mass=None,
                 cell_fields=None, point_fields=None, activity=None,
                 connectivity=None, **kwargs):
        # fill up data objects with obvious data
        pkeys = self.__class__._point_class_._attr_map_
        point_fields = {} if point_fields is None else point_fields
        point_fields[pkeys['loads']] = loads
        point_fields[pkeys['mass']] = mass
        point_fields[pkeys['fixity']] = fixity
        point_fields[pkeys['loads']] = loads
        ckeys = CellData._attr_map_
        cell_fields = {} if cell_fields is None else cell_fields
        cell_fields[ckeys['loads']] = body_loads
        cell_fields[ckeys['strain-loads']] = strain_loads
        cell_fields[ckeys['density']] = density
        cell_fields[ckeys['t']] = t
        cell_fields[ckeys['connectivity']] = connectivity
        super().__init__(*args, point_fields=point_fields,
                         cell_fields=cell_fields, **kwargs)
        # nodal data can only be provided for the root object
        if not self.is_root():
            assert loads is None, "At object creation, nodal loads can only \
                be provided at the top level."
            assert fixity is None, "At object creation, fixity information \
                can only be provided at the top level."
            assert mass is None, "At object creation, nodal masses can only \
                be provided at the top level."
        # it is determined by the size of `activity` whether it refers to
        # cells or points
        if isinstance(activity, np.ndarray):
            nA = activity.shape[0]
            if self.celldata is not None:
                N = len(self.celldata)
                if nA == N:
                    self.celldata.activity = activity
                    nA = -1
            if nA > 0 and self.pointdata is not None:
                N = len(self.pointdata)
                if nA == N:
                    self.pointdata.activity = activity
                    nA = -1
            assert nA < 0
        self._model = model

    def cells_coords(self, *args, points=None, cells=None, **kwargs) \
            -> Union[np.ndarray, List[np.ndarray]]:
        """
        Returns the coordinates of the cells as a 3d numpy array if the 
        subblocks form a regular mesh, or a list of such arrays.
        """
        if points is None and cells is None:
            return super().cells_coords(*args, **kwargs)
        else:
            blocks = self.cellblocks(inclusive=True)
            kwargs.update(points=points, squeeze=False)
            if cells is not None:
                kwargs.update(cells=cells)
                res = {}
                def foo(b): return b.celldata.coords(*args, **kwargs)
                [res.update(d) for d in map(foo, blocks)]
                return res
            else:
                def foo(b): return b.celldata.coords(*args, **kwargs)
                return np.vstack(list(map(foo, blocks)))

    def element_dof_numbering(self, *args, **kwargs) -> np.ndarray:
        """
        Returns global ids of the local degrees of freedoms for each cell.
        """
        blocks = self.cellblocks(inclusive=True)
        def foo(b): return b.celldata.global_dof_numbering()
        return np.vstack(list(map(foo, blocks)))

    def elastic_stiffness_matrix(self, *args, sparse=True, **kwargs) -> Union[np.ndarray, coo_matrix]:
        """
        Returns the elastic stiffness matrix in dense or sparse format.

        Parameters
        ----------
        sparse : bool, Optional
            Returns the matrix in a sparse format. Default is True.

        Returns
        -------
        :class:`numpy.ndarray` or :class:`scipy.sparse.coo_matrix`

        """
        if sparse:
            return self.elastic_stiffness_matrix_coo(*args, **kwargs)
        blocks = self.cellblocks(inclusive=True)
        def foo(b): return b.celldata.elastic_stiffness_matrix()
        return np.vstack(list(map(foo, blocks)))

    def elastic_stiffness_matrix_coo(self, *args, eliminate_zeros=True,
                                     sum_duplicates=True, **kwargs):
        """
        Elastic stiffness matrix in coo format.

        Parameters
        ----------
        eliminate_zeros : bool, Optional
            Eliminates zero entries. Default is True.

        sum_duplicates : bool, Optional
            Sums duplicate entries. Default is True.

        Returns
        -------
        :class:`scipy.sparse.coo_matrix`
            The stiffness matrix in sparse COO format.

        """
        blocks = self.cellblocks(inclusive=True)
        def foo(b): return b.celldata.elastic_stiffness_matrix_coo()
        K = np.sum(list(map(foo, blocks))).tocoo()
        if eliminate_zeros:
            K.eliminate_zeros()
        if sum_duplicates:
            K.sum_duplicates()
        return K

    def masses(self, *args, **kwargs) -> np.ndarray:
        """
        Returns cell masses of the cells as a 1d numpy array.
        """
        blocks = self.cellblocks(*args, inclusive=True, **kwargs)
        vmap = map(lambda b: b.celldata.masses(), blocks)
        return np.concatenate(list(vmap))

    def mass(self, *args, **kwargs) -> float:
        """
        Returns the total mass as a float.
        """
        return np.sum(self.masses(*args, **kwargs))

    def consistent_mass_matrix(self, *args, sparse=True, **kwargs):
        """
        Returns the consistent mass matrix of the mesh with either dense 
        or sparse layout.

        Parameters
        ----------
        sparse : bool, Optional
            Returns the matrix in a sparse format. Default is True.

        Notes
        -----
        If there are nodal masses defined, only sparse output is 
        available at the moment.

        """
        if sparse:
            return self.consistent_mass_matrix_coo(*args, **kwargs)
        else:
            dbkey = self.__class__._point_class_._attr_map_['mass']
            if dbkey in self.root().pointdata.fields:
                return self.consistent_mass_matrix_coo(*args, **kwargs)
        # distributed masses (cells)
        blocks = self.cellblocks(inclusive=True)
        def foo(b): return b.celldata.consistent_mass_matrix()
        return np.vstack(list(map(foo, blocks)))

    def consistent_mass_matrix_coo(self, *args, eliminate_zeros=True,
                                   sum_duplicates=True, distribute=False, **kwargs):
        """
        Returns the mass matrix in coo format. If `distribute` is set 
        to `True`, nodal masses are distributed over neighbouring cells 
        and handled as self-weight is.

        Parameters
        ----------
        eliminate_zeros : bool, Optional
            Eliminates zero entries. Default is True.

        sum_duplicates : bool, Optional
            Sums duplicate entries. Default is True.

        distribute : bool, Optional
            If True, nodal masses are distributed over neighbouring cells 
            and are handled as self-weight is. Default is True.

        Returns
        -------
        scipy.sparse.coo_matrix
            The mass matrix in sparse COO format.

        """
        # distributed masses (cells)
        blocks = list(self.cellblocks(inclusive=True))
        def foo(b): return b.celldata.consistent_mass_matrix_coo()
        M = np.sum(list(map(foo, blocks))).tocoo()
        # nodal masses
        pd = self.root().pointdata
        dbkey = self.__class__._point_class_._attr_map_['mass']
        if dbkey in pd.fields:
            d = pd.mass
            if distribute:
                v = self.volumes()
                edata = list(
                    map(lambda b: b.celldata.pull(data=d, avg=v), blocks))

                def foo(bv): return bv[0].celldata.consistent_mass_matrix_coo(
                    values=bv[1])
                moo = map(lambda i: (blocks[i], edata[i]), range(len(blocks)))
                M += np.sum(list(map(foo, moo))).tocoo()
            else:
                ndof = self.__class__.NDOFN
                M += fem_nodal_mass_matrix_coo(values=d, eliminate_zeros=eliminate_zeros,
                                               sum_duplicates=sum_duplicates, ndof=ndof)
        if eliminate_zeros:
            M.eliminate_zeros()
        if sum_duplicates:
            M.sum_duplicates()
        return M

    def penalty_matrix_coo(self, *args, eliminate_zeros=True,
                           sum_duplicates=True, ensure_comp=False,
                           distribute=False, **kwargs) -> coo_matrix:
        """
        A penalty matrix that enforces Dirichlet boundary conditions. 
        Returns a scipy sparse matrix in coo format.
        """
        fixity = self.root().pointdata.fixity
        K_coo = fem_penalty_matrix_coo(values=fixity, eliminate_zeros=eliminate_zeros,
                                       sum_duplicates=sum_duplicates)
        return K_coo.tocoo()

    def approximation_matrix_coo(self, *args, eliminate_zeros=True, **kwargs) -> coo_matrix:
        blocks = self.cellblocks(inclusive=True)
        def foo(b): return b.celldata.approximation_matrix_coo()
        res = np.sum(list(map(foo, blocks))).tocoo()
        if eliminate_zeros:
            res.eliminate_zeros()
        return res

    def nodal_approximation_matrix_coo(self, *args, eliminate_zeros=True, **kwargs) -> coo_matrix:
        blocks = self.cellblocks(inclusive=True)
        def foo(b): return b.celldata.nodal_approximation_matrix_coo()
        res = np.sum(list(map(foo, blocks))).tocoo()
        if eliminate_zeros:
            res.eliminate_zeros()
        return res

    @squeeze(True)
    def load_vector(self, *args, **kwargs) -> np.ndarray:
        """
        Returns the nodal load vector.

        """
        # concentrated nodal loads
        nodal_data = self.root().pointdata.loads
        nodal_data = atleast3d(nodal_data, back=True)  # (nP, nDOF, nRHS)
        res = fem_load_vector(values=nodal_data, squeeze=False)
        # cells
        blocks = list(self.cellblocks(inclusive=True))
        # body loads
        def foo(b): return b.celldata.body_load_vector(squeeze=False)
        try:
            m = filter(lambda i : i is not None, map(foo, blocks))
            res += np.sum(list(m), axis=0)
        except Exception:
            pass
        # strain loads
        def foo(b): return b.celldata.strain_load_vector(squeeze=False)
        try:
            m = filter(lambda i : i is not None, map(foo, blocks))
            res += np.sum(list(m), axis=0)
        except Exception:
            pass
        return res

    def prostproc_dof_solution(self, *args, **kwargs):
        """
        Calculates approximate solution of the primary variables as a list of arrays
        for each block in the mesh.
        """
        blocks = self.cellblocks(inclusive=True)
        dofsol = self.root().pointdata.dofsol
        def foo(b): return b.celldata.prostproc_dof_solution(dofsol=dofsol)
        list(map(foo, blocks))

    @squeeze(True)
    def nodal_dof_solution(self, *args, flatten=False, **kwargs) -> np.ndarray:
        """
        Returns nodal degree of freedom solution.
        """
        dofsol = self.root().pointdata.dofsol
        if flatten:
            if len(dofsol.shape) == 2:
                return dofsol.flatten()
            else:
                nN, nDOFN, nRHS = dofsol.shape
                return dofsol.reshape((nN * nDOFN, nRHS))
        else:
            return dofsol

    @squeeze(True)
    def cell_dof_solution(self, *args, cells=None, flatten=True,
                          squeeze=True, **kwargs) -> np.ndarray:
        """
        Returns degree of freedom solution for each cell.
        """
        blocks = self.cellblocks(inclusive=True)
        kwargs.update(flatten=flatten, squeeze=False)
        if cells is not None:
            kwargs.update(cells=cells)
            res = {}
            def foo(b): return b.celldata.dof_solution(*args, **kwargs)
            [res.update(d) for d in map(foo, blocks)]
            return res
        else:
            def foo(b): return b.celldata.dof_solution(*args, **kwargs)
            return np.vstack(list(map(foo, blocks)))

    @squeeze(True)
    def strains(self, *args, cells=None, squeeze=True, **kwargs) -> np.ndarray:
        """
        Returns strain solution for each cell.
        """
        blocks = self.cellblocks(inclusive=True)
        kwargs.update(squeeze=False)
        if cells is not None:
            kwargs.update(cells=cells)
            res = {}
            def foo(b): return b.celldata.strains(*args, **kwargs)
            [res.update(d) for d in map(foo, blocks)]
            return res
        else:
            def foo(b): return b.celldata.strains(*args, **kwargs)
            return np.vstack(list(map(foo, blocks)))

    @squeeze(True)
    def internal_forces(self, *args, cells=None, flatten=True, squeeze=True, **kwargs) -> np.ndarray:
        """
        Returns internal force solution for each cell.
        """
        blocks = self.cellblocks(inclusive=True)
        kwargs.update(flatten=flatten, squeeze=False)
        if cells is not None:
            kwargs.update(cells=cells)
            res = {}
            def foo(b): return b.celldata.internal_forces(*args, **kwargs)
            [res.update(d) for d in map(foo, blocks)]
            return res
        else:
            def foo(b): return b.celldata.internal_forces(*args, **kwargs)
            return np.vstack(list(map(foo, blocks)))

    def stresses_at_cells_nodes(self, *args, **kwargs) -> np.ndarray:
        """
        Returns stresses at all nodes of all cells.
        """
        blocks = self.cellblocks(inclusive=True)
        def foo(b): return b.celldata.stresses_at_nodes(*args, **kwargs)
        return np.vstack(list(map(foo, blocks)))

    def stresses_at_centers(self, *args, **kwargs) -> np.ndarray:
        """
        Returns stresses at the centers of all cells.
        """
        blocks = self.cellblocks(inclusive=True)
        def foo(b): return b.celldata.stresses_at_centers(*args, **kwargs)
        return np.squeeze(np.vstack(list(map(foo, blocks))))

    @property
    def model(self):
        """
        Returns the attached model object if there is any.
        """
        if self._model is not None:
            return self._model
        else:
            if self.is_root():
                return self._model
            else:
                return self.parent.model

    def material_stiffness_matrix(self):
        """
        Returns the model of the mesh if there is any.
        This is not important if all the celldata in the mesh
        are filled up with appropriate model stiffness matrices.
        Otherwise, the model stiffness matrix is calcualated on the 
        mesh level during the initialization stage and is spread 
        among the subblocks.
        """
        m = self.model
        if isinstance(m, np.ndarray):
            return m
        else:
            try:
                return m.stiffness_matrix()
            except Exception:
                raise RuntimeError("Invalid model type {}".format(type(m)))

    def postprocess(self, *args, **kwargs):
        """
        General postprocessing. This may be reimplemented in subclasses,
        currently nothing happens here.
        """
        pass
    
    
class SolidMesh(FemMesh):
    """
    A data class dedicated to simple 3d solid cells with 3 degrees of freedom per node.    
    
    See also
    --------
    :class:`TET4`
    :class:`TET10`
    :class:`H8`
    :class:`H27`
    """

    dofs = ('UX', 'UY', 'UZ')
    NDOFN = 3
    
    _cell_classes_ = {
        4: TET4,
        10: TET10,
        8: H8,
        27: H27,
    }

    


# -*- coding: utf-8 -*-
from scipy.sparse import coo_matrix
import numpy as np
from numpy import ndarray
from collections import namedtuple, Iterable
from typing import Iterable, Union
from functools import partial

from neumann import squeeze, config
from neumann.linalg import ReferenceFrame
from neumann.array import atleast1d, atleastnd, ascont
from neumann.utils import to_range
from polymesh.utils import distribute_nodal_data, collect_nodal_data
from polymesh.topo import TopologyArray
from neumann.linalg.sparse.jaggedarray import JaggedArray

from ..preproc import fem_coeff_matrix_coo
from ..postproc import approx_element_solution_bulk, calculate_internal_forces_bulk, \
    explode_kinetic_strains
from ..utils import topo_to_gnum, approximation_matrix, nodal_approximation_matrix, \
    nodal_compatibility_factors, compatibility_factors_to_coo, \
    compatibility_factors, penalty_factor_matrix, assemble_load_vector, \
    tr_cells_1d_in_multi, tr_cells_1d_out_multi, element_dof_solution_bulk, \
    transform_stiffness, constrain_local_stiffness_bulk, assert_min_stiffness_bulk, \
    expand_stiffness_bulk, element_dofmap_bulk, topo_to_gnum_jagged
from ..tr import nodal_dcm, element_dcm

from .meta import FemMixin
from .celldata import CellData
from .utils import stiffness_matrix_bulk2, strain_displacement_matrix_bulk2, \
    unit_strain_load_vector_bulk, strain_load_vector_bulk, mass_matrix_bulk

Quadrature = namedtuple('QuadratureRule', ['inds', 'pos', 'weight'])


UX, UY, UZ, ROTX, ROTY, ROTZ = FX, FY, FZ, MX, MY, MZ = list(range(6))
ulabels = {UX: 'UX', UY: 'UY', UZ: 'UZ',
           ROTX: 'ROTX', ROTY: 'ROTY', ROTZ: 'ROTZ'}
flabels = {FX: 'FX', FY: 'FY', FZ: 'FZ',
           MX: 'MX', MY: 'MY', MZ: 'MZ'}
ulabel_to_int = {v: k for k, v in ulabels.items()}
flabel_to_int = {v: k for k, v in flabels.items()}


class FemCellData(CellData):

    _attr_map_ = {
        'K': 'K',
        'M': 'M',
        'B': 'B',
    }

    @property
    def _dbkey_stiffness_matrix_(self) -> str:
        return self.__class__._attr_map_['K']

    @property
    def _dbkey_strain_displacement_matrix_(self) -> str:
        return self.__class__._attr_map_['B']

    @property
    def _dbkey_mass_matrix_(self) -> str:
        return self.__class__._attr_map_['M']


class FiniteElement(FemCellData, FemMixin):
    """
    Base mixin class for all element types.
    """
    
    def nodal_dcm_matrix(self):
        """
        Returns nodal direction cosine matrices for all elements
        in the block.
            
        Returns
        -------
        numpy.ndarray
            A 3d numpy float array.
        
        See also
        --------
        :func:`..tr.nodal_dcm`
        
        """
        nDOF = self.__class__.NDOFN
        c, r = divmod(nDOF, 3)
        assert r == 0
        frames = self.frames  # dcm_G_L
        return nodal_dcm(frames, c)

    def direction_cosine_matrix(self, *args, source: ndarray = None,
                                target: ndarray = None, N=None, **kwargs):
        """
        Returns the DCM matrix from local to global for all elements 
        in the block.

        Parameters
        ----------
        source : ndarray, Optional
            A source frame. Default is None.

        target : ndarray, Optional
            A target frame. Default is None.

        N : int, Optional
            Number of nodes. If not specified, it is inferred from the class
            of the instance the function is called upon. Default is None.

        Returns
        -------
        ndarray
            The dcm matrix for linear transformations from source to target.

        """
        nNE = self.__class__.NNODE if N is None else N
        nDOF = self.__class__.NDOFN
        ndcm = self.nodal_dcm_matrix()
        dcm = element_dcm(ndcm, nNE, nDOF)
        if source is not None:
            if isinstance(source, str):
                if source == 'global':
                    S = ReferenceFrame(dim=dcm.shape[-1])
                    return ReferenceFrame(dcm).dcm(source=S)
            elif isinstance(source, ReferenceFrame):
                assert source.dim == 2
                return ReferenceFrame(dcm).dcm(source=source)
            else:
                raise NotImplementedError
        elif target is not None:
            if isinstance(target, str):
                if target == 'global':
                    return ReferenceFrame(dcm)
            elif isinstance(target, ReferenceFrame):
                assert target.dim == 2
                return ReferenceFrame(dcm).dcm(target=target)
            else:
                raise NotImplementedError
        return dcm

    def transform_coeff_matrix(self, A, *args, **kwargs):
        """
        Transforms element coefficient matrices (eg. the stiffness or the 
        mass matrix) from local to global.

        Parameters
        ----------
        *args
            Forwarded to :func:`direction_cosine_matrix`

        **kwargs
            Forwarded to :func:`direction_cosine_matrix`

        A : ndarray
            The coefficient matrix in the source frame.

        Returns
        -------
        ndarray
            The coefficient matrix in the target frame.

        """
        dcm = self.direction_cosine_matrix(*args, **kwargs)
        return A if dcm is None else transform_stiffness(A, dcm)

    @squeeze(True)
    def dof_solution(self, *args, target='local', cells=None, points=None,
                     rng=None, flatten=True, **kwargs) -> Union[dict, ndarray]:
        """
        Returns nodal displacements for the cells, wrt. their local frames.

        Parameters
        ----------
        points : float or ndarray, Optional
                Points of evaluation. If provided, it is assumed that the given values
                are wrt. the range [0, 1], unless specified otherwise with the 'rng' parameter. 
                If not provided, results are returned for the nodes of the selected elements.
                Default is None.

        rng : ndarray, Optional
            Range where the points of evauation are understood. Default is [0, 1].

        cells : int or ndarray, Optional
            Indices of cells. If not provided, results are returned for all cells.
            Default is None.

        target : ndarray, Optional
            Reference frame for the output. A value of None or 'local' refers to the local system
            of the cells. Default is 'local'.

        ---
        Returns
        -------
        ndarray or dict
            (nE, nEVAB, nRHS) if flatten else (nE, nNE, nDOF, nRHS) 

        """
        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            if len(cells) == 0:
                return {}
        else:
            cells = np.s_[:]

        dofsol = self.pointdata.dofsol
        dofsol = atleastnd(dofsol, 3, back=True)
        nP, nDOF, nRHS = dofsol.shape
        dofsol = dofsol.reshape(nP * nDOF, nRHS)
        gnum = self.global_dof_numbering()[cells]
        dcm = self.direction_cosine_matrix()[cells]

        # transform values to cell-local frames
        values = element_dof_solution_bulk(dofsol, gnum)  # (nE, nEVAB, nRHS)
        values = ascont(np.swapaxes(values, 1, 2))  # (nE, nRHS, nEVAB)
        values = tr_cells_1d_in_multi(values, dcm)
        values = ascont(np.swapaxes(values, 1, 2))  # (nE, nEVAB, nRHS)

        if points is None:
            points = np.array(self.lcoords()).flatten()
            rng = [-1, 1]
        else:
            rng = np.array([0, 1]) if rng is None else np.array(rng)

        # approximate at points
        # values -> (nE, nEVAB, nRHS)
        points, rng = to_range(
            points, source=rng, target=[-1, 1]).flatten(), [-1, 1]
        N = self.shape_function_matrix(points, rng=rng)[cells]
        # N -> (nE, nP, nDOF, nDOF * nNODE)
        values = ascont(np.swapaxes(values, 1, 2))  # (nE, nRHS, nEVAB)
        values = approx_element_solution_bulk(values, N)
        # values -> (nE, nRHS, nP, nDOF)

        if target is not None:
            # transform values to a destination frame, otherwise return
            # the results in the local frames of the cells
            if isinstance(target, str):
                if target == 'local':
                    pass
                elif target == 'global':
                    nDOF = values.shape[2]
                    target = ReferenceFrame(dim=nDOF)
            if isinstance(target, ReferenceFrame):
                nE, nRHS, nP, nDOF = values.shape
                values = values.reshape(nE, nRHS, nP * nDOF)
                dcm = self.direction_cosine_matrix(N=nP)[cells]
                values = tr_cells_1d_out_multi(values, dcm)
                values = values.reshape(nE, nRHS, nP, nDOF)

        values = np.moveaxis(values, 1, -1)  # (nE, nP, nDOF, nRHS)

        if flatten:
            nE, nP, nDOF, nRHS = values.shape
            values = values.reshape(nE, nP * nDOF, nRHS)

        # values -> (nE, nP, nDOF, nRHS)
        if isinstance(cells, slice):
            # results are requested on all elements
            data = values
        elif isinstance(cells, Iterable):
            data = {c: values[i] for i, c in enumerate(cells)}
        else:
            raise TypeError(
                "Invalid data type <> for cells.".format(type(cells)))

        return data

    @squeeze(True)
    def strains(self, *args, cells=None, points=None, rng=None,
                separate=False, **kwargs) -> Union[dict, ndarray]:
        """
        Returns strains for the cells.

        Parameters
        ----------
        points : float or Iterable[float], Optional
                Points of evaluation. If provided, it is assumed that the given values
                are wrt. the range [0, 1], unless specified otherwise with the 'rng' parameter. 
                If not provided, results are returned for the nodes of the selected elements.
                Default is None.

        rng : Iterable[float], Optional
            Range where the points of evauation are understood. Default is [0, 1].

        cells : int or Iterable[int], Optional.
            Indices of cells. If not provided, results are returned for all cells.
            Default is None.

        separate : bool, Optional
            The different sources are separarated from each other. Default is False.


        Returns
        -------
        ndarray or dict
            (nE, nP * nSTRE, nRHS) if flatten else (nE, nP, nSTRE, nRHS) 

        """
        assert separate is False  # separate heat and force sources

        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            if len(cells) == 0:
                return {}
        else:
            cells = np.s_[:]

        dofsol = self.pointdata.dofsol
        dofsol = atleastnd(dofsol, 3, back=True)
        nP, nDOF, nRHS = dofsol. shape
        dofsol = dofsol.reshape(nP * nDOF, nRHS)
        gnum = self.global_dof_numbering()[cells]
        dcm = self.direction_cosine_matrix()[cells]
        ecoords = self.local_coordinates()[cells]

        # transform values to cell-local frames
        values = element_dof_solution_bulk(dofsol, gnum)  # (nE, nEVAB, nRHS)
        values = ascont(np.swapaxes(values, 1, 2))
        values = tr_cells_1d_in_multi(values, dcm)
        values = ascont(np.swapaxes(values, 1, 2))

        if points is None:
            points = np.array(self.lcoords()).flatten()
            rng = [-1, 1]
        else:
            rng = np.array([0, 1]) if rng is None else np.array(rng)

        # approximate at points
        # values : (nE, nEVAB, nRHS)
        points, rng = to_range(
            points, source=rng, target=[-1, 1]).flatten(), [-1, 1]

        dshp = self.shape_function_derivatives(points, rng=rng)[cells]
        jac = self.jacobian_matrix(
            dshp=dshp, ecoords=ecoords)  # (nE, nP, 1, 1)
        B = self.strain_displacement_matrix(
            dshp=dshp, jac=jac)  # (nE, nP, 4, nNODE * 6)

        values = ascont(np.swapaxes(values, 1, 2))  # (nE, nRHS, nEVAB)
        values = approx_element_solution_bulk(values, B)  # (nE, nRHS, nP, 4)
        values = ascont(np.moveaxis(values, 1, -1))   # (nE, nP, nDOF, nRHS)

        # values : (nE, nP, 4, nRHS)
        if isinstance(cells, slice):
            # results are requested on all elements
            data = values
        elif isinstance(cells, Iterable):
            data = {c: values[i] for i, c in enumerate(cells)}
        else:
            raise TypeError(
                "Invalid data type <> for cells.".format(type(cells)))

        return data

    @squeeze(True)
    def kinetic_strains(self, *args, cells=None, points=None, **kwargs) -> ndarray:
        """
        Returns kinetic strains.

        Parameters
        ----------
        points : scalar or array, Optional
                 Points of evaluation. If provided, it is assumed that the given values
                 are wrt. the range [0, 1], unless specified otherwise with the 'rng' 
                 parameter. If not provided, results are returned for the nodes of the 
                 selected elements. Default is None.

        rng : array, Optional
            Range where the points of evauation are understood. Default is [0, 1].

        cells : integer or array, Optional.
            Indices of cells. If not provided, results are returned for all cells.
            Default is None.

        Returns
        -------
        numpy.ndarray 
            An array of shape (nE, nSTRE, nRHS), where nE is the number of cells,
            nSTRE is the number of strain components and nRHS is the number of 
            load cases.

        """
        key = self.__class__._attr_map_['strain-loads']
        try:
            sloads = self.db[key].to_numpy()
        except Exception as e:
            if key not in self.db.fields:
                nE = len(self)
                nodal_loads = self.pointdata.loads
                if len(nodal_loads.shape) == 2:
                    nRHS = 1
                else:
                    nRHS = self.pointdata.loads.shape[-1]
                nSTRE = self.__class__.NSTRE
                sloads = np.zeros((nE, nSTRE, nRHS))
            else:
                raise e
        finally:
            sloads = atleastnd(sloads, 3, back=True)  # (nE, nSTRE=4, nRHS)
        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            if len(cells) == 0:
                return {}
        else:
            cells = np.s_[:]
        if isinstance(points, Iterable):
            nP = len(points)
            return explode_kinetic_strains(sloads[cells], nP)
        else:
            return sloads[cells]

    @squeeze(True)
    def internal_forces(self, *args, cells=None, points=None, rng=None,
                        flatten=True, target='local', **kwargs) -> Union[dict, ndarray]:
        """
        Returns internal forces for many cells and evaluation points.

        Parameters
        ----------
        points : float or Iterable[float], Optional
                Points of evaluation. If provided, it is assumed that the given values
                are wrt. the range [0, 1], unless specified otherwise with the 'rng' parameter. 
                If not provided, results are returned for the nodes of the selected elements.
                Default is None.

        rng : Iterable[float], Optional
            Range where the points of evauation are understood. 
            Only for 1d cells. Default is [0, 1].

        cells : int or Iterable[int], Optional
            Indices of cells. If not provided, results are returned for all cells.
            Default is None.

        Returns
        -------
        :class:`numpy.ndarray` or :class:`dict`
            (nE, nP * nSTRE, nRHS) if flatten else (nE, nP, nSTRE, nRHS), 
            where nE is the number of cells, nP the number of evaluation points, 
            nSTRE is the number of strain components and nRHS is the number of 
            load cases. 

        """
        if cells is not None:
            cells = atleast1d(cells)
            conds = np.isin(cells, self.id)
            cells = atleast1d(cells[conds])
            if len(cells) == 0:
                return {}
        else:
            cells = np.s_[:]

        dofsol = self.pointdata.dofsol
        dofsol = atleastnd(dofsol, 3, back=True)
        nP, nDOF, nRHS = dofsol. shape
        dofsol = dofsol.reshape(nP * nDOF, nRHS)
        gnum = self.global_dof_numbering()[cells]
        dcm = self.direction_cosine_matrix()[cells]
        ecoords = self.local_coordinates()[cells]

        # transform dofsol to cell-local frames
        dofsol = element_dof_solution_bulk(dofsol, gnum)  # (nE, nEVAB, nRHS)
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))
        dofsol = tr_cells_1d_in_multi(dofsol, dcm)
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))  # (nE, nEVAB, nRHS)

        if points is None:
            points = np.array(self.lcoords()).flatten()
            rng = [-1, 1]
        else:
            if isinstance(points, Iterable):
                points = np.array(points)
            rng = np.array([0, 1]) if rng is None else np.array(rng)

        # approximate at points
        # values : (nE, nEVAB, nRHS)
        points, rng = to_range(
            points, source=rng, target=[-1, 1]).flatten(), [-1, 1]

        dshp = self.shape_function_derivatives(points, rng=rng)[cells]
        jac = self.jacobian_matrix(
            dshp=dshp, ecoords=ecoords)  # (nE, nP, 1, 1)
        B = self.strain_displacement_matrix(
            dshp=dshp, jac=jac)  # (nE, nP, 4, nNODE * 6)

        dofsol = ascont(np.swapaxes(dofsol, 1, 2))  # (nE, nRHS, nEVAB)
        strains = approx_element_solution_bulk(dofsol, B)  # (nE, nRHS, nP, 4)
        strains -= self.kinetic_strains(points=points, squeeze=False)[cells]
        D = self.model_stiffness_matrix()[cells]
        forces = calculate_internal_forces_bulk(
            strains, D)  # (nE, nRHS, nP, 4)
        forces = ascont(np.moveaxis(forces, 1, -1))   # (nE, nP, 4, nRHS)
        dofsol = ascont(np.swapaxes(dofsol, 1, 2))  # (nE, nEVAB, nRHS)
        forces = self._postproc_local_internal_forces(forces, points=points, rng=rng,
                                                      cells=cells, dofsol=dofsol)

        if target is not None:
            # transform values to a destination frame, otherwise return
            # the results in the local frames of the cells
            assert target == 'local'

        if flatten:
            nE, nP, nSTRE, nRHS = forces.shape
            forces = forces.reshape(nE, nP * nSTRE, nRHS)

        # values : (nE, nP, 4, nRHS)
        if isinstance(cells, slice):
            # results are requested on all elements
            data = forces
        elif isinstance(cells, Iterable):
            data = {c: forces[i] for i, c in enumerate(cells)}
        else:
            raise TypeError(
                "Invalid data type <> for cells.".format(type(cells)))

        return data

    def approximation_matrix(self, *args, **kwargs):
        key = self.__class__._attr_map_['ndf']
        return approximation_matrix(self.db[key].to_numpy(), self.NDOFN)

    def approximation_matrix_coo(self, *args, **kwargs):
        N = len(self.pointdata) * self.NDOFN
        topo = self.topology().to_numpy()
        appr = self.approximation_matrix()
        gnum = self.global_dof_numbering(topo=topo)
        return fem_coeff_matrix_coo(appr, *args, inds=gnum, N=N, **kwargs)

    def nodal_approximation_matrix(self, *args, **kwargs):
        key = self.__class__._attr_map_['ndf']
        return nodal_approximation_matrix(self.db[key].to_numpy())

    def nodal_approximation_matrix_coo(self, *args, **kwargs):
        N = len(self.pointdata)
        topo = self.topology().to_numpy()
        appr = self.nodal_approximation_matrix()
        return fem_coeff_matrix_coo(appr, *args, inds=topo, N=N, **kwargs)

    def distribute_nodal_data(self, data, key):
        key = self.__class__._attr_map_['ndf']
        topo = self.topology().to_numpy()
        ndf = self.db[key].to_numpy()
        self.db[key] = distribute_nodal_data(data, topo, ndf)

    def collect_nodal_data(self, key, *args, N=None, **kwargs):
        topo = self.topology().to_numpy()
        N = len(self.pointdata) if N is None else N
        cellloads = self.db[key].to_numpy()
        return collect_nodal_data(cellloads, topo, N)

    def compatibility_penalty_matrix_coo(self, *args, nam_csr_tot, p=1e12, **kwargs):
        """
        Returns a matrix that penalizes incompatible displacement modes.

        Parameters
        ----------
        nam_csr_tot : nodal_approximation matrix for the whole structure
                      in csr format

        p : float, Optional
            Courant-type penalty parameter. Default is 1e12.

        """
        topo = self.topology().to_numpy()
        nam = self.nodal_approximation_matrix()
        factors, nreg = nodal_compatibility_factors(nam_csr_tot, nam, topo)
        factors, nreg = compatibility_factors(factors, nreg, self.NDOFN)
        data, rows, cols = compatibility_factors_to_coo(factors, nreg)
        return coo_matrix((data*p, (rows, cols)))

    def penalty_factor_matrix(self, *args, **kwargs):
        cellfixity = self.fixity
        shp = self.shape_function_values(self.lcoords())
        return penalty_factor_matrix(cellfixity, shp)

    def penalty_matrix(self, *args, p=1e12, **kwargs):
        return self.penalty_factor_matrix() * p

    def penalty_matrix_coo(self, *args, p=1e12, **kwargs):
        nP = len(self.pointdata)
        N = nP * self.NDOFN
        topo = self.topology().to_numpy()
        K_bulk = self.penalty_matrix(*args, topo=topo, p=p, **kwargs)
        gnum = self.global_dof_numbering(topo=topo)
        return fem_coeff_matrix_coo(K_bulk, *args, inds=gnum, N=N, **kwargs)

    def global_dof_numbering(self, *args, **kwargs) -> Union[TopologyArray, ndarray]:
        """
        Returns global numbering of the degrees of freedom of the cells in the block.        
        """
        topo = kwargs.get("topo", None)
        topo = self.topology() if topo is None else topo
        nDOFN = self.container.NDOFN
        try:
            if topo.is_jagged():
                cuts = topo.widths() * nDOFN
                data1d = np.zeros(np.sum(cuts), dtype=int)
                gnum = JaggedArray(data1d, cuts=cuts)
                topo_to_gnum_jagged(topo, gnum, nDOFN)
                return TopologyArray(gnum, to_numpy=False)
            else:
                return topo_to_gnum(topo.to_numpy(), nDOFN)
        except Exception:
            return topo_to_gnum(topo, nDOFN)

    @squeeze(True)
    def elastic_stiffness_matrix(self, *args, transform=True, **kwargs) -> np.ndarray:
        """
        Returns the elastic stiffness matrix of the cells of the instance.

        Parameters
        ----------
        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.

        Returns
        -------
        :class:`numpy.ndarray`
            The stiffness matrices as a 3d numpy array, where the elements run along
            the first axis.

        """
        # build
        K = self._elastic_stiffness_matrix_(*args, transform=False, **kwargs)

        # handle connectivity
        try:
            conn = self.connectivity
            if isinstance(conn, ndarray):
                # only for line meshes
                if len(conn.shape) == 3 and conn.shape[1] == 2:  # nE, 2, nDOF
                    nE, _, nDOF = conn.shape
                    factors = np.ones((nE, K.shape[-1]))
                    factors[:, :nDOF] = conn[:, 0, :]
                    factors[:, -nDOF:] = conn[:, -1, :]
                    nEVAB2 = 2 * nDOF - 0.5
                    cond = np.sum(factors, axis=1) < nEVAB2
                    i = np.where(cond)[0]
                    K[i] = constrain_local_stiffness_bulk(K[i], factors[i])
                    assert_min_stiffness_bulk(K)
                else:
                    raise NotImplementedError(
                        "Unknown shape of <{}> for 'connectivity'.".format(conn.shape))
        except Exception:
            pass
        
        # store
        dbkey = self._dbkey_stiffness_matrix_
        self.db[dbkey] = K

        # expand, transform, return
        nDOFN = self.container.NDOFN
        dofmap = self.__class__.dofmap
        if len(dofmap) < nDOFN:
            nE = K.shape[0]
            nX = nDOFN * self.NNODE
            K_ = np.zeros((nE, nX, nX), dtype=float)
            dofmap = element_dofmap_bulk(dofmap, nDOFN, self.NNODE)
            K = expand_stiffness_bulk(K, K_, dofmap)
            assert_min_stiffness_bulk(K)

        return self.transform_coeff_matrix(K) if transform else K

    @config(store_strains=False)
    def _elastic_stiffness_matrix_(self, *args, transform: bool = True, **kwargs):

        ec = kwargs.get('_ec', None)
        if ec is None:
            nSTRE = self.__class__.NSTRE
            nDOF = self.__class__.NDOFN
            nNE = self.__class__.NNODE
            nE = len(self)

            _topo = kwargs.get('_topo', self.topology().to_numpy())
            _frames = kwargs.get('_frames', self.frames)
            if _frames is not None:
                ec = self.local_coordinates(_topo=_topo, frames=_frames)
            else:
                ec = self.points_of_cells(topo=_topo)

            if kwargs.get('store_strains', False):
                dbkey = self._dbkey_strain_displacement_matrix_
                self.db[dbkey] = np.zeros((nE, nSTRE, nDOF * nNE))

        q = kwargs.get('_q', None)
        if q is None:
            # main loop
            D = self.model_stiffness_matrix()
            func = partial(self._elastic_stiffness_matrix_, _D=D, **kwargs)
            q = self.quadrature[self.qrule]
            N = nDOF * nNE
            K = np.zeros((len(self), N, N))
            if isinstance(q, dict):
                # many side loops
                for qinds, qvalue in q.items():
                    if isinstance(qvalue, str):
                        qpos, qweight = self.quadrature[qvalue]
                    else:
                        qpos, qweight = qvalue
                    q = Quadrature(qinds, qpos, qweight)
                    K += func(_q=q, _ec=ec)
            else:
                # one side loop
                qpos, qweight = self.quadrature[self.qrule]
                q = Quadrature(None, qpos, qweight)
                K = func(_q=q, _ec=ec)
            # end of main cycle
            dbkey = self._dbkey_stiffness_matrix_
            self.db[dbkey] = K
            return self.transform_coeff_matrix(K) if transform else K

        # in side loop
        dshp = self.shape_function_derivatives(q.pos)
        jac = self.jacobian_matrix(dshp=dshp, ecoords=ec)
        B = self.strain_displacement_matrix(dshp=dshp, jac=jac)
        if q.inds is not None:
            # zero out unused indices, only for selective integration
            inds = np.where(~np.in1d(np.arange(self.NSTRE), q.inds))[0]
            B[:, :, inds, :] = 0.
        djac = self.jacobian(jac=jac)
        if kwargs.get('store_strains', False):
            # B (nE, nG, nSTRE=4, nNODE * nDOF=6)
            dbkey = self._dbkey_strain_displacement_matrix_
            _B = self.db[dbkey].to_numpy()
            _B += strain_displacement_matrix_bulk2(B, djac, q.weight)
            self.db[dbkey] = _B
        D = kwargs.get('_D', self.model_stiffness_matrix())
        return stiffness_matrix_bulk2(D, B, djac, q.weight)

    def elastic_stiffness_matrix_coo(self, *args, **kwargs) -> coo_matrix:
        """
        Returns the elastic stiffness matrix of the cells of the instance
        in sparse COO format.

        Returns
        -------
        coo_matrix
            The stiffness matrix as a 2d sparse array in COO format.

        """
        nP = len(self.pointdata)
        N = nP * self.NDOFN
        topo = self.topology().to_numpy()
        K_bulk = self.elastic_stiffness_matrix(*args, _topo=topo, **kwargs)
        gnum = self.global_dof_numbering(topo=topo)
        return fem_coeff_matrix_coo(K_bulk, *args, inds=gnum, N=N, **kwargs)

    @squeeze(True)
    def consistent_mass_matrix(self, *args, **kwargs) -> ndarray:
        """
        Returns the mass matrix of the cells of the instance.

        Returns
        -------
        :class:`ndarray`
            The mass matrices as a 3d numpy array, where the elements run along
            the first axis.

        """
        return self._consistent_mass_matrix_(*args, **kwargs)

    def _consistent_mass_matrix_(self, *args, values=None, **kwargs):
        _topo = kwargs.get('_topo', self.topology().to_numpy())
        if 'frames' not in kwargs:
            key = self.__class__._attr_map_['frames']
            if key in self.db.fields:
                frames = self.frames
            else:
                frames = None
        _ecoords = kwargs.get('_ecoords', None)
        if _ecoords is None:
            if frames is not None:
                _ecoords = self.local_coordinates(_topo=_topo, frames=frames)
            else:
                _ecoords = self.points_of_cells(topo=_topo)
        if isinstance(values, np.ndarray):
            _dens = values
        else:
            _dens = kwargs.get('_dens', self.density)

        try:
            _areas = self.areas()
        except Exception:
            _areas = np.ones_like(_dens)

        q = kwargs.get('_q', None)
        if q is None:
            q = self.quadrature['mass']
            if isinstance(q, dict):
                N = self.NNODE * self.NDOFN
                nE = len(self)
                res = np.zeros((nE, N, N))
                for qinds, qvalue in q.items():
                    if isinstance(qvalue, str):
                        qpos, qweight = self.quadrature[qvalue]
                    else:
                        qpos, qweight = qvalue
                    q = Quadrature(qinds, qpos, qweight)
                    res += self._consistent_mass_matrix_(*args, _topo=_topo,
                                                         frames=frames, _q=q,
                                                         values=_dens,
                                                         _ecoords=_ecoords,
                                                         **kwargs)
            else:
                qpos, qweight = self.quadrature['mass']
                q = Quadrature(None, qpos, qweight)
                res = self._consistent_mass_matrix_(*args, _topo=_topo,
                                                    frames=frames, _q=q,
                                                    values=_dens,
                                                    _ecoords=_ecoords,
                                                    **kwargs)

            key = self.__class__._attr_map_['M']
            self.db[key] = res
            return self.transform_coeff_matrix(res)

        rng = np.array([-1., 1.])
        dshp = self.shape_function_derivatives(q.pos)
        jac = self.jacobian_matrix(dshp=dshp, ecoords=_ecoords)
        djac = self.jacobian(jac=jac)
        N = self.shape_function_matrix(q.pos, rng=rng)
        return mass_matrix_bulk(N, _dens, _areas, djac, q.weight)

    def consistent_mass_matrix_coo(self, *args, **kwargs):
        """
        Returns the mass matrix of the cells of the instance
        in sparse COO format.

        Returns
        -------
        coo_matrix
            The stiffness matrix as a 2d sparse array in COO format.

        """
        nP = len(self.pointdata)
        N = nP * self.NDOFN
        topo = self.topology().to_numpy()
        M_bulk = self.consistent_mass_matrix(*args, **kwargs)
        gnum = self.global_dof_numbering(topo=topo)
        return fem_coeff_matrix_coo(M_bulk, *args, inds=gnum, N=N, **kwargs)

    @squeeze(True)
    def strain_load_vector(self, values=None, *args, squeeze, 
                           return_zeroes=False, **kwargs) -> ndarray:
        """
        Generates a load vector from strain loads specified for all cells.

        Parameters
        ----------
        values : ndarray, Optional
            Strain loads as a 3d numpy array of shape (nE, nS, nRHS). 
            The array must contain values for all cells (nE), strain 
            components (nS) and load cases (nL).
            
        return_zeroes : bool, Optional
            Controls what happends if there are no strain loads provided.
            If True, a zero array is retured with correct shape, otherwise None. 
            Default is False.

        Returns
        -------
        ndarray
            The equivalent load vector.

        """
        nRHS = self.pointdata.loads.shape[-1]
        nSTRE = self.__class__.NSTRE
        dbkey = self.__class__._attr_map_['strain-loads']
        try:
            if values is None:
                values = self.db[dbkey].to_numpy()
        except Exception as e:
            if dbkey not in self.db.fields:
                if not return_zeroes:
                    return None
                nE = len(self)
                values = np.zeros((nE, nSTRE, nRHS))
            else:
                raise e
        finally:
            values = atleastnd(values, 3, back=True)  # (nE, nSTRE=4, nRHS)

        dbkey = self.__class__._attr_map_['B']
        if dbkey not in self.db.fields:
            self.elastic_stiffness_matrix(*args, squeeze=False,
                                          store_strains=True, **kwargs)

        B = self.db[dbkey].to_numpy()  # (nE, nSTRE=4, nNODE * nDOF=6)
        D = self.model_stiffness_matrix()  # (nE, nSTRE=4, nSTRE=4)
        BTD = unit_strain_load_vector_bulk(D, B)
        values = np.swapaxes(values, 1, 2)  # (nE, nRHS, nSTRE=4)
        nodal_loads = strain_load_vector_bulk(BTD, ascont(values))
        return self.transform_local_nodal_loads(nodal_loads)

    @squeeze(True)
    def body_load_vector(self, values=None, *args, source=None,
                         constant=False, target=None, 
                         return_zeroes=False, **kwargs) -> ndarray:
        """
        Builds the equivalent nodal node representation of body loads
        and returns it in the global frame, or an arbitrary target frame 
        if specified.

        Parameters
        ----------
        values : ndarray, Optional
            Body load values for all cells. Default is None.

        source : ndarray, Optional
            The frame in which the provided values are to be understood.
            Default is None.

        constant : bool, Optional
            ...

        target : ndarray, Optional
            The frame in which the load vectors are expected.
            Default is None.
            
        return_zeroes : bool, Optional
            Controls what happends if there are no strain loads provided.
            If True, a zero array is retured with correct shape, otherwise None. 
            Default is False.

        Returns
        -------
        ndarray
            The nodal load vector for all load cases as a 2d numpy array.

        """
        nRHS = self.pointdata.loads.shape[-1]
        nNE = self.__class__.NNODE
        dbkey = self.__class__._attr_map_['loads']
        try:
            if values is None:
                values = self.db[dbkey].to_numpy()
        except Exception as e:
            if dbkey not in self.db.fields:
                if not return_zeroes:
                    return None
                nE = len(self)
                values = np.zeros((nE, nNE, nDOF, nRHS))
            else:
                raise e
        finally:
            values = atleastnd(values, 4, back=True)  # (nE, nNE, nDOF, nRHS)

        # prepare data to shape (nE, nNE * nDOF, nRHS)
        if constant:
            values = atleastnd(values, 3, back=True)  # (nE, nDOF, nRHS)
            #np.insert(values, 1, values, axis=1)
            nE, nDOF, nRHS = values.shape
            values_ = np.zeros((nE, nNE, nDOF, nRHS), dtype=values.dtype)
            for i in range(nNE):
                values_[:, i, :, :] = values
            values = values_
        values = atleastnd(values, 4, back=True)  # (nE, nNE, nDOF, nRHS)
        nE, _, nDOF, nRHS = values.shape
        # (nE, nNE, nDOF, nRHS) -> (nE, nNE * nDOF, nRHS)
        values = values.reshape(nE, nNE * nDOF, nRHS)
        values = ascont(values)

        if source is not None:
            raise NotImplementedError
            nE, nNE, nDOF, nRHS = values.shape
            dcm = self.direction_cosine_matrix(source=source)
            # (nE, nNE * nDOF, nRHS) -> (nE, nRHS, nNE * nDOF)
            values = np.swapaxes(values, 1, 2)
            values = ascont(values)
            values = tr_cells_1d_in_multi(values, dcm)
            # (nE, nRHS, nNE * nDOF) -> (nE, nNE * nDOF, nRHS)
            values = np.swapaxes(values, 1, 2)
            values = ascont(values)

        nodal_loads = self.integrate_body_loads(values)
        return self.transform_local_nodal_loads(nodal_loads)

    def transform_local_nodal_loads(self, nodal_loads: ndarray):
        """
        nodal_loads (nE, nNE * nDOF, nRHS) == (nE, nX, nRHS)
        (nX, nRHS)
        """
        dcm = self.direction_cosine_matrix()
        # (nE, nNE * nDOF, nRHS) -> (nE, nRHS, nNE * nDOF)
        nodal_loads = np.swapaxes(nodal_loads, 1, 2)
        nodal_loads = ascont(nodal_loads)
        nodal_loads = tr_cells_1d_out_multi(nodal_loads, dcm)
        # (nE, nRHS, nNE * nDOF) -> (nE, nNE * nDOF, nRHS)
        nodal_loads = np.swapaxes(nodal_loads, 1, 2)
        # assemble
        topo = self.topology().to_numpy()
        gnum = self.global_dof_numbering(topo=topo)
        nX = len(self.pointdata) * self.container.NDOFN
        return assemble_load_vector(nodal_loads, gnum, nX)  # (nX, nRHS)
